# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""controlled_noisy_web_labels dataset."""

import csv
import dataclasses
import io
import json
import os

from etils import epath
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
Controlled Noisy Web Labels is a collection of ~212,000 URLs to images in which
every image is carefully annotated by 3-5 labeling professionals by Google Cloud
Data Labeling Service. Using these annotations, it establishes the first
benchmark of controlled real-world label noise from the web.

We provide the Red Mini-ImageNet (real-world web noise) and Blue Mini-ImageNet
configs:
  - controlled_noisy_web_labels/mini_imagenet_red
  - controlled_noisy_web_labels/mini_imagenet_blue

Each config contains ten variants with ten noise-levels p from 0% to 80%. The
validation set has clean labels and is shared across all noisy training sets.
Therefore, each config has the following splits:

  - train_00
  - train_05
  - train_10
  - train_15
  - train_20
  - train_30
  - train_40
  - train_50
  - train_60
  - train_80
  - validation

The details for dataset construction and analysis can be found in the paper.
All images are resized to 84x84 resolution.

"""

_CITATION = """
@inproceedings{jiang2020beyond,
  title={Beyond synthetic noise: Deep learning on controlled noisy labels},
  author={Jiang, Lu and Huang, Di and Liu, Mason and Yang, Weilong},
  booktitle={International Conference on Machine Learning},
  pages={4804--4815},
  year={2020},
  organization={PMLR}
}
"""

_PERCENTS = [0, 5, 10, 15, 20, 30, 40, 50, 60, 80]

# n01440764 is not part of mini-ImageNet but we add it here for testing
_IMNET_MINI_SYNSETS = [
    'n01704323',
    'n04515003',
    'n02101006',
    'n03062245',
    'n04509417',
    'n03854065',
    'n02110063',
    'n03998194',
    'n02111277',
    'n04443257',
    'n02165456',
    'n03075370',
    'n02747177',
    'n01558993',
    'n04149813',
    'n03220513',
    'n03584254',
    'n04243546',
    'n03770439',
    'n02108915',
    'n02971356',
    'n02108551',
    'n03908618',
    'n01981276',
    'n03535780',
    'n04275548',
    'n03272010',
    'n02138441',
    'n03337140',
    'n07584110',
    'n01910747',
    'n03146219',
    'n02443484',
    'n04146614',
    'n02113712',
    'n02981792',
    'n04251144',
    'n03838899',
    'n04604644',
    'n02116738',
    'n02099601',
    'n02966193',
    'n02687172',
    'n03924679',
    'n02120079',
    'n03476684',
    'n04435653',
    'n02091244',
    'n01843383',
    'n01532829',
    'n03980874',
    'n07613480',
    'n03544143',
    'n03347037',
    'n02114548',
    'n06794110',
    'n04258138',
    'n02606052',
    'n02105505',
    'n02871525',
    'n03127925',
    'n04418357',
    'n04389033',
    'n02074367',
    'n02110341',
    'n13133613',
    'n03773504',
    'n13054560',
    'n07747607',
    'n04522168',
    'n02089867',
    'n02129165',
    'n02457408',
    'n03676483',
    'n09256479',
    'n04067472',
    'n02823428',
    'n01749939',
    'n03400231',
    'n04596742',
    'n01855672',
    'n02174001',
    'n04612504',
    'n03047690',
    'n02091831',
    'n03017168',
    'n04296562',
    'n02108089',
    'n09246464',
    'n02219486',
    'n02950826',
    'n01770081',
    'n03775546',
    'n01930112',
    'n03888605',
    'n03417042',
    'n03527444',
    'n02795169',
    'n07697537',
    'n03207743',
    'n01440764',
]

MINI_IMAGENET_TRAIN = 'https://raw.githubusercontent.com/yaoyao-liu/mini-imagenet-tools/main/csv_files/train.csv'
MINI_IMAGENET_VAL = 'https://raw.githubusercontent.com/yaoyao-liu/mini-imagenet-tools/main/csv_files/val.csv'
MINI_IMAGENET_TEST = 'https://raw.githubusercontent.com/yaoyao-liu/mini-imagenet-tools/main/csv_files/test.csv'


@dataclasses.dataclass
class ControlledNoisyWebLabelsConfig(tfds.core.BuilderConfig):
  name: str = ''
  dataset: str = 'mini_imagenet'
  color: str = ''
  num_classes: int = 100


class ControlledNoisyWebLabels(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for controlled_noisy_web_labels dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  In order to manually download this data, a user must perform the
  following operations:
  
  1. Download the splits and the annotations [here](https://storage.googleapis.com/cnlw/dataset_no_images.zip)
  2. Extract dataset_no_images.zip to dataset_no_images/.
  3. Download all images in dataset_no_images/mini-imagenet-annotations.json into
  a new folder named dataset_no_images/noisy_images/. The output filename must
  agree with the image id provided in mini-imagenet-annotations.json. For example,
  if "image/id": "5922767e5677aef4", then the downloaded image should be
  dataset_no_images/noisy_images/5922767e5677aef4.jpg.
  4.Register on https://image-net.org/download-images and download
  ILSVRC2012_img_train.tar and ILSVRC2012_img_val.tar.

  The resulting directory structure may then be processed by TFDS:

  - dataset_no_images/
    - mini-imagenet/
      - class_name.txt
      - split/
        - blue_noise_nl_0.0
        - blue_noise_nl_0.1
        - ...
        - red_noise_nl_0.0
        - red_noise_nl_0.1
        - ...
        - clean_validation
    - mini-imagenet-annotations.json
  - ILSVRC2012_img_train.tar
  - ILSVRC2012_img_val.tar
  - noisy_images/
    - 5922767e5677aef4.jpg

  """
  # pytype: disable=wrong-keyword-args
  BUILDER_CONFIGS = [
      ControlledNoisyWebLabelsConfig(
          name='mini_imagenet_red',
          dataset='mini_imagenet',
          color='red',
          num_classes=100,
      ),
      ControlledNoisyWebLabelsConfig(
          name='mini_imagenet_blue',
          dataset='mini_imagenet',
          color='blue',
          num_classes=100,
      ),
  ]

  # pytype: enable=wrong-keyword-args

  def _read_mini_imagenet_csv(self, path):
    fnames = []
    with tf.io.gfile.GFile(path, 'r') as f:
      reader = csv.reader(f)
      next(reader, None)
      for data in reader:
        fnames.append(data[0])
    return fnames

  def _resize_image(self, buf, resolution=84):
    cv2 = tfds.core.lazy_imports.cv2

    image = cv2.imdecode(np.frombuffer(buf, dtype=np.uint8), flags=3)
    im_resized = cv2.resize(
        image, (resolution, resolution), interpolation=cv2.INTER_AREA
    )
    _, buff = cv2.imencode(
        '.jpg', im_resized, [int(cv2.IMWRITE_JPEG_QUALITY), 72]
    )
    return buff.tobytes()

  def _get_clean_images(self, fnames, archive):
    clean_image_dict = {}

    fnames = set(fnames)

    # The files names in mini-ImageNet contain the absolute index wrt to the
    # synset. For example, {synset_name}_{idx}.jpg, where synset_name is
    # something like n01440764, and idx is between 1 and 1000. We want to create
    # a mapping between synset and the indices of the images that we need for
    # the corresponding synset
    synset_to_index = {}
    for name in fnames:
      # Remove .jpg
      name = name[:-4]
      # Synset name is the first 8 characters
      synset_name = name[:9]
      # Synset index is the last 8 characters
      synset_index = int(name[-4:])

      if synset_name in synset_to_index:
        synset_to_index[synset_name].append(synset_index)
      else:
        synset_to_index[synset_name] = [synset_index]

    for fname, fobj in archive:
      label = fname[:-4]  # fname is something like 'n01632458.tar'
      # TODO(b/117643231): in py3, the following lines trigger tarfile module
      # to call `fobj.seekable()`, which Gfile doesn't have. We should find an
      # alternative, as this loads ~150MB in RAM.

      if label not in _IMNET_MINI_SYNSETS or label not in synset_to_index:
        continue
      fobj_mem = io.BytesIO(fobj.read())

      # We obtain all filenames for a given subset (i.e. n01440764_10,
      # n0144076_2020) and sort them, so that we can find the indices of images
      # that we need to keep
      all_image_names = [
          image_fname
          for image_fname, _ in tfds.download.iter_archive(
              fobj_mem, tfds.download.ExtractMethod.TAR_STREAM
          )
      ]

      # Remove .JPEG, and remove the part after _, i.e. want to retain
      # 1 from n01440764_1
      image_indices = [
          int(name[:-5].split('_')[-1]) for name in all_image_names
      ]

      # Sort filenames
      sorted_indices = np.argsort(image_indices)

      # Get the indices of images to select
      images_to_select = np.asarray(synset_to_index[label]) - 1

      # Get the filenames of images to select
      selected_indices = sorted_indices[images_to_select]
      selected_fnames = [all_image_names[idx] for idx in selected_indices]
      fname_to_idx = dict(zip(selected_fnames, synset_to_index[label]))

      # We use set here for faster lookup in the loop
      selected_fnames = set(selected_fnames)

      # integer label corresponding to the synset
      synset_int_label = _IMNET_MINI_SYNSETS.index(label)

      # Go to the beginning of the file
      fobj.seek(0)
      fobj_mem = io.BytesIO(fobj.read())

      for image_fname, image in tfds.download.iter_archive(
          fobj_mem, tfds.download.ExtractMethod.TAR_STREAM
      ):  # pytype: disable=wrong-arg-types  # gen-stub-imports
        if image_fname in selected_fnames:
          # Convert the image ID back to mini-Imagenet style, e.g.
          # n0144076400000009.jpg, for a quick access later
          image_index = str(fname_to_idx[image_fname]).zfill(8)
          new_image_id = label + image_index + '.jpg'
          buf = image.read()
          clean_image_dict[new_image_id] = (
              self._resize_image(buf),
              synset_int_label,
          )

    return clean_image_dict

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    features = {}
    features['image'] = tfds.features.Image()
    features['label'] = tfds.features.ClassLabel(
        num_classes=self.builder_config.num_classes
    )
    features['is_clean'] = tfds.features.Tensor(shape=(), dtype=np.bool_)
    features['id'] = tfds.features.Text()

    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(features),
        # If there's a common (input, target) tuple from the
        # features, specify them here. They'll be used if
        # `as_supervised=True` in `builder.as_dataset`.
        supervised_keys=('image', 'label'),  # Set to `None` to disable
        homepage=(
            'https://google.github.io/controlled-noisy-web-labels/index.html'
        ),
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    imnet_path = os.path.join(dl_manager.manual_dir, 'ILSVRC2012_img_train.tar')
    noisy_images_path = os.path.join(dl_manager.manual_dir, 'noisy_images')
    noisy_split_path = os.path.join(
        dl_manager.manual_dir, 'dataset_no_images', 'mini-imagenet', 'split'
    )
    noisy_annot_path = os.path.join(
        dl_manager.manual_dir,
        'dataset_no_images',
        'mini-imagenet-annotations.json',
    )
    val_path = os.path.join(dl_manager.manual_dir, 'ILSVRC2012_img_val.tar')

    with epath.Path(noisy_annot_path).open() as json_file:
      data = json.load(json_file)

    image_data = data['data']
    noisy_image_ids = [elem[0]['image/id'] + '.jpg' for elem in image_data]

    # We first load all mini-ImageNet images in the memory, and
    # will access them for the other splits
    paths = dl_manager.download({
        'mini_train': MINI_IMAGENET_TRAIN,
        'mini_val': MINI_IMAGENET_VAL,
        'mini_test': MINI_IMAGENET_TEST,
    })

    train_fnames = self._read_mini_imagenet_csv(paths['mini_train'])
    val_fnames = self._read_mini_imagenet_csv(paths['mini_val'])
    test_fnames = self._read_mini_imagenet_csv(paths['mini_test'])
    mini_imnet_fnames = train_fnames + val_fnames + test_fnames

    mini_imnet_images = self._get_clean_images(
        mini_imnet_fnames, dl_manager.iter_archive(imnet_path)
    )

    val_split_file = os.path.join(noisy_split_path, 'clean_validation')

    split_to_generator = {}

    split_to_generator[tfds.Split.VALIDATION] = self._generate_val_examples(
        val_split_file, dl_manager.iter_archive(val_path)
    )

    for percent in _PERCENTS:
      split_name = tfds.Split.TRAIN + '_' + '{:02d}'.format(percent)
      split_file = os.path.join(
          noisy_split_path,
          '{}_noise_nl_{}'.format(
              self.builder_config.color, str(percent / 100)
          ),
      )
      split_to_generator[split_name] = self._generate_examples(
          split_file, noisy_image_ids, noisy_images_path, mini_imnet_images
      )

    return split_to_generator

  def _generate_val_examples(self, split_file, archive):
    """Yields examples for the validation set."""
    with epath.Path(split_file).open() as fp:
      split_info = fp.read().splitlines()

    split_names, split_labels = map(
        list, zip(*[elem.split(' ') for elem in split_info])
    )

    for fname, fobj in archive:
      # Remove the .JPEG
      fname = fname[:-5]
      if fname + '.jpg' in split_names:
        label_idx = split_names.index(fname + '.jpg')
        example = {}
        example['is_clean'] = 1
        buf = fobj.read()
        example['image'] = self._resize_image(buf)
        example['label'] = int(split_labels[label_idx])
        example['id'] = fname

        yield fname, example

  def _generate_examples(
      self, split_file, noisy_image_ids, noisy_images_path, clean_images
  ):
    """Yields examples."""

    with epath.Path(split_file).open() as fp:
      split_info = fp.read().splitlines()

    split_names, split_labels = map(
        list, zip(*[elem.split(' ') for elem in split_info])
    )

    # We use a set here because searching in a set is faster than a list
    noisy_image_ids = set(noisy_image_ids)

    for imname, label in zip(split_names, split_labels):
      example = {}
      if imname in noisy_image_ids:
        # If the image comes from noise image list, it's not clean
        example['is_clean'] = 0
        with tf.io.gfile.GFile(
            os.path.join(noisy_images_path, imname), 'rb'
        ) as f:
          example['image'] = self._resize_image(f.read())
      else:
        example['image'] = clean_images[imname][0]
        # Check if the assigned label and the real label is the same
        if clean_images[imname][1] == int(label):
          example['is_clean'] = 1
        else:
          example['is_clean'] = 0

      example['label'] = int(label)
      example['id'] = imname[:-4]
      yield imname[:-4], example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Dump the list of all registered datasets/config/version in a `.txt` file.

Instructions:

```
python tensorflow_datasets/scripts/freeze_dataset_version.py
```


"""

from absl import app

import tensorflow_datasets as tfds


def main(_):
  tfds.core.visibility.set_availables([
      tfds.core.visibility.DatasetType.TFDS_PUBLIC,
  ])

  registered_names = tfds.core.load.list_full_names()
  version_path = tfds.core.utils.tfds_write_path() / 'stable_versions.txt'
  version_path.write_text('\n'.join(registered_names))
  print(f'{len(registered_names)} datasets versions written.')


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Graph visualizer."""

from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Union

from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.core.visualization import visualizer

_GraphFn = Callable[[Any], Any]
_Color = Union[float, str]
_NodeColorFn = Callable[[Any], _Color]
_EdgeColorFn = Callable[[Any], _Color]
_NodeColorDict = Dict[Any, _Color]
_EdgeColorDict = Dict[Any, _Color]


class GraphVisualizerMetadataDict(dataset_info.MetadataDict):
  """A `tfds.core.Metadata` object for metadata for visualizing graph datasets.

  This metadata information is used by tfds.core.visualization.GraphVisualizer.
  """

  def __init__(self, edgelist_feature_name, *args, **kwargs):
    super(GraphVisualizerMetadataDict, self).__init__(*args, **kwargs)
    self['edgelist_feature_name'] = edgelist_feature_name


def _extract_metadata_dict(
    metadata: Optional[dataset_info.Metadata],
) -> Optional[GraphVisualizerMetadataDict]:
  """Extracts out the metadata dict for the GraphVisualizer."""
  if metadata is None:
    return None

  for metadata_entry in metadata.values():
    for key in metadata_entry:
      if key == 'edgelist_feature_name':
        return metadata_entry
  return None


class GraphVisualizer(visualizer.Visualizer):
  """Visualizer for graph prediction datasets."""

  def match(self, ds_info: dataset_info.DatasetInfo) -> bool:
    """Checks whether this dataset can be visualized with this visualizer.

    See base class for more information.
    Args:
      ds_info: Metadata for the dataset.

    Returns:
      boolean value indicating whether the current visualizer can be used.
    """
    graph_viz_metadata = _extract_metadata_dict(ds_info.metadata)
    return graph_viz_metadata is not None

  def show(
      self,
      ds: tf.data.Dataset,
      ds_info: dataset_info.DatasetInfo,
      node_label_fn: Optional[_GraphFn] = None,
      node_color_fn: Optional[_GraphFn] = None,
      edge_color_fn: Optional[_GraphFn] = None,
      rows: int = 3,
      cols: int = 3,
      plot_scale: float = 5,
      **kwargs,
  ):
    """Display the dataset.

    Args:
      ds: `tf.data.Dataset`. The tf.data.Dataset object to visualize. Examples
        should not be batched. Examples will be consumed in order until (rows *
        cols) are read or the dataset is consumed.
      ds_info: `tfds.core.DatasetInfo` object of the dataset to visualize.
      node_label_fn: A callable that maps individual graph examples to a
        dictionary of node labels, rendered within the nodes.
      node_color_fn: A callable that maps individual graph examples to a
        dictionary of node colors.
      edge_color_fn: A callable that maps individual graph examples to a
        dictionary of edge colors.
      rows: `int`, number of rows of the display grid.
      cols: `int`, number of columns of the display grid.
      plot_scale: `float`, controls the plot size of the images. Keep this value
        around 5 to get a good plot. High and low values may cause the labels to
        get overlapped.
      **kwargs: Additional arguments passed to networkx.draw_networkx().

    Returns:
      fig: The pyplot figure.
    """

    plt = lazy_imports_lib.lazy_imports.matplotlib.pyplot
    nx = lazy_imports_lib.lazy_imports.networkx

    # Construct edge-getter function which will extract edges from each graph.
    get_edges_fn = _make_edge_getter_fn(ds_info)

    # Create subplots.
    fig, axs = plt.subplots(
        nrows=rows,
        ncols=cols,
        squeeze=False,
        figsize=(plot_scale * cols, plot_scale * rows),
    )
    plt.subplots_adjust(hspace=1 / plot_scale, wspace=1 / plot_scale)

    for graph, ax in zip(ds, axs.reshape(-1)):
      # Obtain edges.
      edges = get_edges_fn(graph)

      # Create NetworkX graph.
      nx_graph = nx.Graph()
      nx_graph.add_edges_from(edges)

      # Add node labels, node colors and edge colors, if specified.
      if node_label_fn is not None:
        kwargs['labels'] = node_label_fn(graph)

      if node_color_fn is not None:
        node_color_map = node_color_fn(graph)
        node_color_mapper = _make_node_color_mapper(node_color_map)
        kwargs['node_color'] = list(map(node_color_mapper, nx_graph.nodes()))

      if edge_color_fn is not None:
        edge_color_map = edge_color_fn(graph)
        edge_color_mapper = _make_edge_color_mapper(edge_color_map)
        kwargs['edge_color'] = list(map(edge_color_mapper, nx_graph.edges()))

      # Draw with NetworkX.
      nx.draw_networkx(nx_graph, ax=ax, **kwargs)

    return fig


def _make_edge_getter_fn(ds_info: dataset_info.DatasetInfo) -> _GraphFn:
  """Returns a function which will extract edges from each graph."""
  graph_viz_metadata = _extract_metadata_dict(ds_info.metadata)
  edgelist_feature_name = graph_viz_metadata['edgelist_feature_name']

  def get_edges_fn(graph):
    return graph[edgelist_feature_name].numpy()

  return get_edges_fn


def _make_node_color_mapper(
    node_color_map: _NodeColorDict, default_color: str = 'C0'
) -> _NodeColorFn:
  """Helper to map nodes to colors."""

  def node_color_mapper(node):
    return node_color_map.get(node, default_color)

  return node_color_mapper


def _make_edge_color_mapper(
    edge_color_map: _EdgeColorDict, default_color: str = 'C0'
) -> _EdgeColorFn:
  """Helper to map edges to colors."""

  def edge_color_mapper(edge):
    reversed_edge = tuple(reversed(edge))
    if edge in edge_color_map:
      return edge_color_map[edge]
    elif reversed_edge in edge_color_map:
      return edge_color_map[reversed_edge]
    else:
      return default_color

  return edge_color_mapper



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feature.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rfeature.proto\x12\x13tensorflow_datasets"\xa0\x01\n\x0c\x46\x65\x61turesDict\x12\x41\n\x08\x66\x65\x61tures\x18\x01'
    b' \x03(\x0b\x32/.tensorflow_datasets.FeaturesDict.FeaturesEntry\x1aM\n\rFeaturesEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12+\n\x05value\x18\x02'
    b' \x01(\x0b\x32\x1c.tensorflow_datasets.Feature:\x02\x38\x01"\xbf\x05\n\x07\x46\x65\x61ture\x12\x19\n\x11python_class_name\x18\x01'
    b' \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x0e'
    b' \x01(\t\x12\x13\n\x0bvalue_range\x18\x0f'
    b' \x01(\t\x12\x38\n\x0cjson_feature\x18\x02 \x01(\x0b\x32'
    b' .tensorflow_datasets.JsonFeatureH\x00\x12:\n\rfeatures_dict\x18\x03'
    b' \x01(\x0b\x32!.tensorflow_datasets.FeaturesDictH\x00\x12\x34\n\x06tensor\x18\x04'
    b' \x01(\x0b\x32".tensorflow_datasets.TensorFeatureH\x00\x12\x36\n\x0b\x63lass_label\x18\x05'
    b' \x01(\x0b\x32\x1f.tensorflow_datasets.ClassLabelH\x00\x12\x32\n\x05image\x18\x06'
    b' \x01(\x0b\x32!.tensorflow_datasets.ImageFeatureH\x00\x12\x32\n\x05video\x18\x07'
    b' \x01(\x0b\x32!.tensorflow_datasets.VideoFeatureH\x00\x12\x32\n\x05\x61udio\x18\x08'
    b' \x01(\x0b\x32!.tensorflow_datasets.AudioFeatureH\x00\x12?\n\x0c\x62ounding_box\x18\t'
    b" \x01(\x0b\x32'.tensorflow_datasets.BoundingBoxFeatureH\x00\x12\x30\n\x04text\x18\n"
    b' \x01(\x0b\x32'
    b' .tensorflow_datasets.TextFeatureH\x00\x12>\n\x0btranslation\x18\x0b'
    b" \x01(\x0b\x32'.tensorflow_datasets.TranslationFeatureH\x00\x12\x31\n\x08sequence\x18\x0c"
    b' \x01(\x0b\x32\x1d.tensorflow_datasets.SequenceH\x00\x42\t\n\x07\x63ontent"\x1b\n\x0bJsonFeature\x12\x0c\n\x04json\x18\x01'
    b' \x01(\t"\x1b\n\x05Shape\x12\x12\n\ndimensions\x18\x01'
    b' \x03(\x03"[\n\rTensorFeature\x12)\n\x05shape\x18\x01'
    b' \x01(\x0b\x32\x1a.tensorflow_datasets.Shape\x12\r\n\x05\x64type\x18\x02'
    b' \x01(\t\x12\x10\n\x08\x65ncoding\x18\x03'
    b' \x01(\t"!\n\nClassLabel\x12\x13\n\x0bnum_classes\x18\x01'
    b' \x01(\x03"\xa7\x01\n\x0cImageFeature\x12)\n\x05shape\x18\x01'
    b' \x01(\x0b\x32\x1a.tensorflow_datasets.Shape\x12\r\n\x05\x64type\x18\x02'
    b' \x01(\t\x12\x17\n\x0f\x65ncoding_format\x18\x03'
    b' \x01(\t\x12\x14\n\x0cuse_colormap\x18\x04'
    b' \x01(\x08\x12.\n\x05label\x18\x05'
    b' \x01(\x0b\x32\x1f.tensorflow_datasets.ClassLabel"\x92\x01\n\x0cVideoFeature\x12)\n\x05shape\x18\x01'
    b' \x01(\x0b\x32\x1a.tensorflow_datasets.Shape\x12\r\n\x05\x64type\x18\x02'
    b' \x01(\t\x12\x17\n\x0f\x65ncoding_format\x18\x03'
    b' \x01(\t\x12\x14\n\x0cuse_colormap\x18\x04'
    b' \x01(\x08\x12\x19\n\x11\x66\x66mpeg_extra_args\x18\x05'
    b' \x03(\t"\x99\x01\n\x0c\x41udioFeature\x12)\n\x05shape\x18\x01'
    b' \x01(\x0b\x32\x1a.tensorflow_datasets.Shape\x12\r\n\x05\x64type\x18\x02'
    b' \x01(\t\x12\x13\n\x0b\x66ile_format\x18\x03'
    b' \x01(\t\x12\x13\n\x0bsample_rate\x18\x04'
    b' \x01(\x03\x12\x10\n\x08\x65ncoding\x18\x05'
    b' \x01(\t\x12\x13\n\x0blazy_decode\x18\x06'
    b' \x01(\x08"N\n\x12\x42oundingBoxFeature\x12)\n\x05shape\x18\x01'
    b' \x01(\x0b\x32\x1a.tensorflow_datasets.Shape\x12\r\n\x05\x64type\x18\x02'
    b' \x01(\t"\r\n\x0bTextFeature"O\n\x12TranslationFeature\x12\x11\n\tlanguages\x18\x01'
    b' \x03(\t\x12&\n\x1evariable_languages_per_example\x18\x02'
    b' \x01(\x08"I\n\x08Sequence\x12-\n\x07\x66\x65\x61ture\x18\x01'
    b' \x01(\x0b\x32\x1c.tensorflow_datasets.Feature\x12\x0e\n\x06length\x18\x02'
    b' \x01(\x03\x42\x03\xf8\x01\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feature_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _FEATURESDICT_FEATURESENTRY._options = None
  _FEATURESDICT_FEATURESENTRY._serialized_options = b'8\001'
  _FEATURESDICT._serialized_start = 39
  _FEATURESDICT._serialized_end = 199
  _FEATURESDICT_FEATURESENTRY._serialized_start = 122
  _FEATURESDICT_FEATURESENTRY._serialized_end = 199
  _FEATURE._serialized_start = 202
  _FEATURE._serialized_end = 905
  _JSONFEATURE._serialized_start = 907
  _JSONFEATURE._serialized_end = 934
  _SHAPE._serialized_start = 936
  _SHAPE._serialized_end = 963
  _TENSORFEATURE._serialized_start = 965
  _TENSORFEATURE._serialized_end = 1056
  _CLASSLABEL._serialized_start = 1058
  _CLASSLABEL._serialized_end = 1091
  _IMAGEFEATURE._serialized_start = 1094
  _IMAGEFEATURE._serialized_end = 1261
  _VIDEOFEATURE._serialized_start = 1264
  _VIDEOFEATURE._serialized_end = 1410
  _AUDIOFEATURE._serialized_start = 1413
  _AUDIOFEATURE._serialized_end = 1566
  _BOUNDINGBOXFEATURE._serialized_start = 1568
  _BOUNDINGBOXFEATURE._serialized_end = 1646
  _TEXTFEATURE._serialized_start = 1648
  _TEXTFEATURE._serialized_end = 1661
  _TRANSLATIONFEATURE._serialized_start = 1663
  _TRANSLATIONFEATURE._serialized_end = 1742
  _SEQUENCE._serialized_start = 1744
  _SEQUENCE._serialized_end = 1817
# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""mrqa datasets."""

# pylint: disable=g-complex-comprehension, anomalous-backslash-in-string

from __future__ import annotations

import json
import textwrap

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
The MRQA 2019 Shared Task focuses on generalization in question answering. An
effective question answering system should do more than merely interpolate from
the training set to answer test examples drawn from the same distribution: it
should also be able to extrapolate to out-of-distribution examples — a
significantly harder challenge.

MRQA adapts and unifies multiple distinct question answering datasets (carefully
selected subsets of existing datasets) into the same format (SQuAD format).
Among them, six datasets were made available for training, and six datasets were
made available for testing. Small portions of the training datasets
were held-out as in-domain data that may be used for development. The testing
datasets only contain out-of-domain data. This benchmark is released as part of
the MRQA 2019 Shared Task.

More information can be found at: `https://mrqa.github.io/2019/shared.html`.
"""

_HOMEPAGE = 'https://mrqa.github.io/2019/shared.html'

_CITATION = """@inproceedings{fisch-etal-2019-mrqa,
    title = "{MRQA} 2019 Shared Task: Evaluating Generalization in Reading Comprehension",
    author = "Fisch, Adam  and
      Talmor, Alon  and
      Jia, Robin  and
      Seo, Minjoon  and
      Choi, Eunsol  and
      Chen, Danqi",
    booktitle = "Proceedings of the 2nd Workshop on Machine Reading for Question Answering",
    month = nov,
    year = "2019",
    address = "Hong Kong, China",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/D19-5801",
    doi = "10.18653/v1/D19-5801",
    pages = "1--13",
}

Note that each MRQA dataset has its own citation. Please see the source to see
the correct citation for each contained dataset."
"""

_URL_BASE = 'https://s3.us-east-2.amazonaws.com/mrqa/release/v2'
_URLs = {
    # Train sub-datasets
    'train+SQuAD': f'{_URL_BASE}/train/SQuAD.jsonl.gz',
    'train+NewsQA': f'{_URL_BASE}/train/NewsQA.jsonl.gz',
    'train+TriviaQA': f'{_URL_BASE}/train/TriviaQA-web.jsonl.gz',
    'train+SearchQA': f'{_URL_BASE}/train/SearchQA.jsonl.gz',
    'train+HotpotQA': f'{_URL_BASE}/train/HotpotQA.jsonl.gz',
    'train+NaturalQuestions': (
        f'{_URL_BASE}/train/NaturalQuestionsShort.jsonl.gz'
    ),
    # Validation sub-datasets
    'validation+SQuAD': f'{_URL_BASE}/dev/SQuAD.jsonl.gz',
    'validation+NewsQA': f'{_URL_BASE}/dev/NewsQA.jsonl.gz',
    'validation+TriviaQA': f'{_URL_BASE}/dev/TriviaQA-web.jsonl.gz',
    'validation+SearchQA': f'{_URL_BASE}/dev/SearchQA.jsonl.gz',
    'validation+HotpotQA': f'{_URL_BASE}/dev/HotpotQA.jsonl.gz',
    'validation+NaturalQuestions': (
        f'{_URL_BASE}/dev/NaturalQuestionsShort.jsonl.gz'
    ),
    # Test sub-datasets
    'test+BioASQ': (
        'http://participants-area.bioasq.org/MRQA2019/'
    ),  # BioASQ.jsonl.gz
    'test+DROP': f'{_URL_BASE}/dev/DROP.jsonl.gz',
    'test+DuoRC': f'{_URL_BASE}/dev/DuoRC.ParaphraseRC.jsonl.gz',
    'test+RACE': f'{_URL_BASE}/dev/RACE.jsonl.gz',
    'test+RelationExtraction': f'{_URL_BASE}/dev/RelationExtraction.jsonl.gz',
    'test+TextbookQA': f'{_URL_BASE}/dev/TextbookQA.jsonl.gz',
}


class MRQAConfig(tfds.core.BuilderConfig):
  """BuilderConfig for MRQA."""

  def __init__(self, *, features, data_urls, citation, **kwargs):
    """BuilderConfig for MRQA.

    Args:
      features: `tfds.features.FeaturesDict`, specific feature dict for the
        dataset.
      data_urls: `dict`, urls to download the files from
      citation: `string`, citation for the data set
      **kwargs: keyword arguments forwarded to super.
    """
    super(MRQAConfig, self).__init__(**kwargs)
    self.features = features
    self.data_urls = data_urls
    self.citation = citation


class MRQA(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for mrqa dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  DEFAULT_FEATURES = tfds.features.FeaturesDict({
      'subset': np.str_,
      'context': np.str_,
      'context_tokens': tfds.features.Sequence({
          'tokens': np.str_,
          'offsets': np.int32,
      }),
      'qid': np.str_,
      'question': np.str_,
      'question_tokens': tfds.features.Sequence({
          'tokens': np.str_,
          'offsets': np.int32,
      }),
      'detected_answers': tfds.features.Sequence({
          'text': np.str_,
          'char_spans': tfds.features.Sequence({
              'start': np.int32,
              'end': np.int32,
          }),
          'token_spans': tfds.features.Sequence({
              'start': np.int32,
              'end': np.int32,
          }),
      }),
      'answers': tfds.features.Sequence(np.str_),
  })
  BUILDER_CONFIGS = [
      MRQAConfig(
          name='squad',
          description=textwrap.dedent(
              """\
          The SQuAD (Stanford Question Answering Dataset) dataset is used as the
          basis for the shared task format. Crowdworkers are shown paragraphs
          from Wikipedia and are asked to write questions with extractive
          answers.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': _URLs['train+SQuAD'],
              'validation': _URLs['validation+SQuAD'],
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{rajpurkar-etal-2016-squad,
              title = "{SQ}u{AD}: 100,000+ Questions for Machine Comprehension of Text",
              author = "Rajpurkar, Pranav  and
                Zhang, Jian  and
                Lopyrev, Konstantin  and
                Liang, Percy",
              booktitle = "Proceedings of the 2016 Conference on Empirical Methods in Natural Language Processing",
              month = nov,
              year = "2016",
              address = "Austin, Texas",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D16-1264",
              doi = "10.18653/v1/D16-1264",
              pages = "2383--2392",
          }
          """
          ),
      ),
      MRQAConfig(
          name='news_qa',
          description=textwrap.dedent(
              """\
          Two sets of crowdworkers ask and answer questions based on CNN news
          articles. The “questioners” see only the article’s headline and
          summary while the “answerers” see the full article. Questions that
          have no answer or are flagged in the dataset to be without annotator
          agreement are discarded.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': _URLs['train+NewsQA'],
              'validation': _URLs['validation+NewsQA'],
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{trischler-etal-2017-newsqa,
              title = "{N}ews{QA}: A Machine Comprehension Dataset",
              author = "Trischler, Adam  and
                Wang, Tong  and
                Yuan, Xingdi  and
                Harris, Justin  and
                Sordoni, Alessandro  and
                Bachman, Philip  and
                Suleman, Kaheer",
              booktitle = "Proceedings of the 2nd Workshop on Representation Learning for {NLP}",
              month = aug,
              year = "2017",
              address = "Vancouver, Canada",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/W17-2623",
              doi = "10.18653/v1/W17-2623",
              pages = "191--200",
          }
      #     """
          ),
      ),
      MRQAConfig(
          name='trivia_qa',
          description=textwrap.dedent(
              """\
          Question and answer pairs are sourced from trivia and quiz-league
          websites. The web version of TriviaQA, where the contexts are
          retrieved from the results of a Bing search query, is used.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': _URLs['train+TriviaQA'],
              'validation': _URLs['validation+TriviaQA'],
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{joshi-etal-2017-triviaqa,
              title = "{T}rivia{QA}: A Large Scale Distantly Supervised Challenge Dataset for Reading Comprehension",
              author = "Joshi, Mandar  and
                Choi, Eunsol  and
                Weld, Daniel  and
                Zettlemoyer, Luke",
              booktitle = "Proceedings of the 55th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers)",
              month = jul,
              year = "2017",
              address = "Vancouver, Canada",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/P17-1147",
              doi = "10.18653/v1/P17-1147",
              pages = "1601--1611",
          }
          """
          ),
      ),
      MRQAConfig(
          name='search_qa',
          description=textwrap.dedent(
              """\
          Question and answer pairs are sourced from the Jeopardy! TV show. The
          contexts are composed of retrieved snippets from a Google search
          query.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': _URLs['train+SearchQA'],
              'validation': _URLs['validation+SearchQA'],
          },
          citation=textwrap.dedent(
              """\
          @article{dunn2017searchqa,
              title={Searchqa: A new q\\&a dataset augmented with context from a search engine},
              author={Dunn, Matthew and Sagun, Levent and Higgins, Mike and Guney, V Ugur and Cirik, Volkan and Cho, Kyunghyun},
              journal={arXiv preprint arXiv:1704.05179},
              year={2017}
          }
          """
          ),
      ),
      MRQAConfig(
          name='hotpot_qa',
          description=textwrap.dedent(
              """\
          Crowdworkers are shown two entity-linked paragraphs from Wikipedia and
          are asked to write and answer questions that require multi-hop
          reasoning to solve. In the original setting, these paragraphs are
          mixed with additional distractor paragraphs to make inference harder.
          Here, the distractor paragraphs are not included.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': _URLs['train+HotpotQA'],
              'validation': _URLs['validation+HotpotQA'],
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{yang-etal-2018-hotpotqa,
              title = "{H}otpot{QA}: A Dataset for Diverse, Explainable Multi-hop Question Answering",
              author = "Yang, Zhilin  and
                Qi, Peng  and
                Zhang, Saizheng  and
                Bengio, Yoshua  and
                Cohen, William  and
                Salakhutdinov, Ruslan  and
                Manning, Christopher D.",
              booktitle = "Proceedings of the 2018 Conference on Empirical Methods in Natural Language Processing",
              month = oct # "-" # nov,
              year = "2018",
              address = "Brussels, Belgium",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D18-1259",
              doi = "10.18653/v1/D18-1259",
              pages = "2369--2380",
          }
          """
          ),
      ),
      MRQAConfig(
          name='natural_questions',
          description=textwrap.dedent(
              """\
          Questions are collected from information-seeking queries to the Google
          search engine by real users under natural conditions. Answers to the
          questions are annotated in a retrieved Wikipedia page by crowdworkers.
          Two types of annotations are collected: 1) the HTML bounding box
          containing enough information to completely infer the answer to the
          question (Long Answer), and 2) the subspan or sub-spans within the
          bounding box that comprise the actual answer (Short Answer). Only the
          examples that have short answers are used, and the long answer is used
          as the context.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'train': _URLs['train+NaturalQuestions'],
              'validation': _URLs['validation+NaturalQuestions'],
          },
          citation=textwrap.dedent(
              """\
          @article{kwiatkowski-etal-2019-natural,
              title = "Natural Questions: A Benchmark for Question Answering Research",
              author = "Kwiatkowski, Tom  and
                Palomaki, Jennimaria  and
                Redfield, Olivia  and
                Collins, Michael  and
                Parikh, Ankur  and
                Alberti, Chris  and
                Epstein, Danielle  and
                Polosukhin, Illia  and
                Devlin, Jacob  and
                Lee, Kenton  and
                Toutanova, Kristina  and
                Jones, Llion  and
                Kelcey, Matthew  and
                Chang, Ming-Wei  and
                Dai, Andrew M.  and
                Uszkoreit, Jakob  and
                Le, Quoc  and
                Petrov, Slav",
              journal = "Transactions of the Association for Computational Linguistics",
              volume = "7",
              year = "2019",
              address = "Cambridge, MA",
              publisher = "MIT Press",
              url = "https://aclanthology.org/Q19-1026",
              doi = "10.1162/tacl_a_00276",
              pages = "452--466",
          }
          """
          ),
      ),
      MRQAConfig(
          name='bio_asq',
          description=textwrap.dedent(
              """\
          BioASQ, a challenge on large-scale biomedical semantic indexing and
          question answering, contains question and answer pairs that are
          created by domain experts. They are then manually linked to multiple
          related science (PubMed) articles. The full abstract of each of the
          linked articles is downloaded and used as individual contexts (e.g., a
          single question can be linked to multiple, independent articles to
          create multiple QA-context pairs). Abstracts that do not exactly
          contain the answer are discarded.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'test': _URLs['test+BioASQ'],
          },
          citation=textwrap.dedent(
              """\
          @article{tsatsaronis2015overview,
              title={An overview of the BIOASQ large-scale biomedical semantic indexing and question answering competition},
              author={Tsatsaronis, George and Balikas, Georgios and Malakasiotis, Prodromos and Partalas, Ioannis and Zschunke, Matthias and Alvers, Michael R and Weissenborn, Dirk and Krithara, Anastasia and Petridis, Sergios and Polychronopoulos, Dimitris and others},
              journal={BMC bioinformatics},
              volume={16},
              number={1},
              pages={1--28},
              year={2015},
              publisher={Springer}
          }
          """
          ),
      ),
      MRQAConfig(
          name='drop',
          description=textwrap.dedent(
              """\
          DROP (Discrete Reasoning Over the content of Paragraphs) examples were
          collected similarly to SQuAD, where crowdworkers are asked to create
          question-answer pairs from Wikipedia paragraphs. The questions focus
          on quantitative reasoning, and the original dataset contains
          non-extractive numeric answers as well as extractive text answers. The
          set of questions that are extractive is used.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'test': _URLs['test+DROP'],
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{dua-etal-2019-drop,
              title = "{DROP}: A Reading Comprehension Benchmark Requiring Discrete Reasoning Over Paragraphs",
              author = "Dua, Dheeru  and
                Wang, Yizhong  and
                Dasigi, Pradeep  and
                Stanovsky, Gabriel  and
                Singh, Sameer  and
                Gardner, Matt",
              booktitle = "Proceedings of the 2019 Conference of the North {A}merican Chapter of the Association for Computational Linguistics: Human Language Technologies, Volume 1 (Long and Short Papers)",
              month = jun,
              year = "2019",
              address = "Minneapolis, Minnesota",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/N19-1246",
              doi = "10.18653/v1/N19-1246",
              pages = "2368--2378",
          }
          """
          ),
      ),
      MRQAConfig(
          name='duo_rc',
          description=textwrap.dedent(
              """\
          The ParaphraseRC split of the DuoRC dataset is used. In this setting,
          two different plot summaries of the same movie are collected—one from
          Wikipedia and the other from IMDb. Two different sets of crowdworkers
          ask and answer questions about the movie plot, where the “questioners”
          are shown only the Wikipedia page, and the “answerers” are shown only
          the IMDb page. Questions that are marked as unanswerable are
          discarded.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'test': _URLs['test+DuoRC'],
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{saha-etal-2018-duorc,
              title = "{D}uo{RC}: Towards Complex Language Understanding with Paraphrased Reading Comprehension",
              author = "Saha, Amrita  and
                Aralikatte, Rahul  and
                Khapra, Mitesh M.  and
                Sankaranarayanan, Karthik",
              booktitle = "Proceedings of the 56th Annual Meeting of the Association for Computational Linguistics (Volume 1: Long Papers)",
              month = jul,
              year = "2018",
              address = "Melbourne, Australia",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/P18-1156",
              doi = "10.18653/v1/P18-1156",
              pages = "1683--1693",
          }
          """
          ),
      ),
      MRQAConfig(
          name='race',
          description=textwrap.dedent(
              """\
          ReAding Comprehension Dataset From Examinations (RACE) is collected
          from English reading comprehension exams for middle and high school
          Chinese students. The high school split (which is more challenging)
          is used and also the implicit “fill in the blank” style questions
          (which are unnatural for this task) are filtered out.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'test': _URLs['test+RACE'],
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{lai-etal-2017-race,
              title = "{RACE}: Large-scale {R}e{A}ding Comprehension Dataset From Examinations",
              author = "Lai, Guokun  and
                Xie, Qizhe  and
                Liu, Hanxiao  and
                Yang, Yiming  and
                Hovy, Eduard",
              booktitle = "Proceedings of the 2017 Conference on Empirical Methods in Natural Language Processing",
              month = sep,
              year = "2017",
              address = "Copenhagen, Denmark",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/D17-1082",
              doi = "10.18653/v1/D17-1082",
              pages = "785--794",
          }
          """
          ),
      ),
      MRQAConfig(
          name='relation_extraction',
          description=textwrap.dedent(
              """\
          Given a slot-filling dataset, relations among entities are
          systematically transformed into questionanswer pairs using templates.
          For example, the educated_at(x, y) relationship between two entities x
          and y appearing in a sentence can be expressed as “Where was x
          educated at?” with answer y. Multiple templates for each type of
          relation are collected. The dataset’s zeroshot benchmark split
          (generalization to unseen relations) is used, and only the positive
          examples are kept.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'test': _URLs['test+RelationExtraction'],
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{levy-etal-2017-zero,
              title = "Zero-Shot Relation Extraction via Reading Comprehension",
              author = "Levy, Omer  and
                Seo, Minjoon  and
                Choi, Eunsol  and
                Zettlemoyer, Luke",
              booktitle = "Proceedings of the 21st Conference on Computational Natural Language Learning ({C}o{NLL} 2017)",
              month = aug,
              year = "2017",
              address = "Vancouver, Canada",
              publisher = "Association for Computational Linguistics",
              url = "https://aclanthology.org/K17-1034",
              doi = "10.18653/v1/K17-1034",
              pages = "333--342",
          }
          """
          ),
      ),
      MRQAConfig(
          name='textbook_qa',
          description=textwrap.dedent(
              """\
          TextbookQA is collected from lessons from middle school Life Science,
          Earth Science, and Physical Science textbooks. Questions that are
          accompanied with a diagram, or that are “True or False” questions are
          not included.
          """
          ),
          features=DEFAULT_FEATURES,
          data_urls={
              'test': _URLs['test+TextbookQA'],
          },
          citation=textwrap.dedent(
              """\
          @inproceedings{kembhavi2017you,
              title={Are you smarter than a sixth grader? textbook question answering for multimodal machine comprehension},
              author={Kembhavi, Aniruddha and Seo, Minjoon and Schwenk, Dustin and Choi, Jonghyun and Farhadi, Ali and Hajishirzi, Hannaneh},
              booktitle={Proceedings of the IEEE Conference on Computer Vision and Pattern recognition},
              pages={4999--5007},
              year={2017}
          }
          """
          ),
      ),
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(self.builder_config.features),
        supervised_keys=None,
        homepage=_HOMEPAGE,
        citation=self.builder_config.citation + '\n' + _CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    data_dir = dl_manager.download_and_extract(self.builder_config.data_urls)
    split_generators = {}

    if 'train' in self.builder_config.data_urls:
      split_generators.update(
          {
              tfds.Split.TRAIN: self._generate_examples(
                  path=data_dir['train'], split='train'
              ),
          }
      )

    if 'validation' in self.builder_config.data_urls:
      split_generators.update(
          {
              tfds.Split.VALIDATION: self._generate_examples(
                  path=data_dir['validation'], split='validation'
              ),
          }
      )

    if 'test' in self.builder_config.data_urls:
      split_generators.update(
          {
              tfds.Split.TEST: self._generate_examples(
                  path=data_dir['test'], split='test'
              ),
          }
      )
    return split_generators

  def _generate_examples(self, path, split):
    """Yields examples."""
    with epath.Path(path).open() as f:
      header = next(f)
      subset = json.loads(header)['header']['dataset']

      for row in f:
        paragraph = json.loads(row)
        context = paragraph['context'].strip()
        context_tokens = [
            {'tokens': t[0], 'offsets': t[1]}
            for t in paragraph['context_tokens']
        ]
        for qa in paragraph['qas']:
          qid = qa['qid']
          question = qa['question'].strip()
          question_tokens = [
              {'tokens': t[0], 'offsets': t[1]} for t in qa['question_tokens']
          ]
          detected_answers = []
          for detect_ans in qa['detected_answers']:
            detected_answers.append({
                'text': detect_ans['text'].strip(),
                'char_spans': [
                    {'start': t[0], 'end': t[1]}
                    for t in detect_ans['char_spans']
                ],
                'token_spans': [
                    {'start': t[0], 'end': t[1]}
                    for t in detect_ans['token_spans']
                ],
            })
          answers = qa['answers']
          yield f'{split}+{subset}_{qid}', {
              'subset': subset,
              'context': context,
              'context_tokens': context_tokens,
              'qid': qid,
              'question': question,
              'question_tokens': question_tokens,
              'detected_answers': detected_answers,
              'answers': answers,
          }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Download manager interface."""
from __future__ import annotations

import concurrent.futures
import dataclasses
import functools
import hashlib
import typing
from typing import Any, Dict, Iterator, Optional, Tuple, Union
import uuid

from absl import logging
from etils import epath
import promise

from tensorflow_datasets.core import utils
from tensorflow_datasets.core.download import checksums
from tensorflow_datasets.core.download import downloader
from tensorflow_datasets.core.download import extractor
from tensorflow_datasets.core.download import kaggle
from tensorflow_datasets.core.download import resource as resource_lib
from tensorflow_datasets.core.download import util
from tensorflow_datasets.core.utils import shard_utils
from tensorflow_datasets.core.utils import tree_utils
from tensorflow_datasets.core.utils import type_utils

# pylint: disable=logging-fstring-interpolation

Tree = type_utils.Tree

Url = Union[str, resource_lib.Resource]
ExtractPath = Union[epath.PathLike, resource_lib.Resource]


def get_downloader(*args: Any, **kwargs: Any):
  return downloader.get_downloader(*args, **kwargs)


class NonMatchingChecksumError(Exception):
  """The downloaded file doesn't have expected checksum."""


# Even if `DownloadConfig` is immutable inside TFDS, we do not set `frozen=True`
# to allow user to set:
# dl_config = tfds.download.DownloadConfig()
# dl_config.beam_runner = ...
@dataclasses.dataclass(eq=False)
class DownloadConfig:
  """Configuration for `tfds.core.DatasetBuilder.download_and_prepare`.

  Attributes:
    extract_dir: `str`, directory where extracted files are stored. Defaults to
      "<download_dir>/extracted".
    manual_dir: `str`, read-only directory where manually downloaded/extracted
      data is stored. Defaults to `<download_dir>/manual`.
    download_mode: `tfds.GenerateMode`, how to deal with downloads or data that
      already exists. Defaults to `REUSE_DATASET_IF_EXISTS`, which will reuse
      both downloads and data if it already exists.
    compute_stats: `tfds.download.ComputeStats`, whether to compute statistics
      over the generated data. Defaults to `AUTO`.
    max_examples_per_split: `int`, optional max number of examples to write into
      each split (used for testing). If set to 0, only execute the
      `_split_generators` (download the original data), but skip
      `_generator_examples`.
    register_checksums: `bool`, defaults to False. If True, checksum of
      downloaded files are recorded.
    force_checksums_validation: `bool`, defaults to False. If True, raises an
      error if an URL do not have checksums.
    beam_runner: Runner to pass to `beam.Pipeline`, only used for datasets based
      on Beam for the generation.
    beam_options: `PipelineOptions` to pass to `beam.Pipeline`, only used for
      datasets based on Beam for the generation.
    try_download_gcs: `bool`, defaults to True. If True, prepared dataset will
      be downloaded from GCS, when available. If False, dataset will be
      downloaded and prepared from scratch.
    verify_ssl: `bool`, defaults to True. If True, will verify certificate when
      downloading dataset.
    override_max_simultaneous_downloads: `int`, optional max number of
      simultaneous downloads. If set, it will override dataset builder and
      downloader default values.
    num_shards: optional number of shards that should be created. If `None`,
      then the number of shards is computed based on the total size of the
      dataset and the min and max shard size.
    min_shard_size: optional minimum shard size in bytes. If `None`, 64 MB is
      used.
    max_shard_size: optional maximum shard size in bytes. If `None`, 1 GiB is
      used.
  """

  extract_dir: Optional[epath.PathLike] = None
  manual_dir: Optional[epath.PathLike] = None
  download_mode: util.GenerateMode = util.GenerateMode.REUSE_DATASET_IF_EXISTS
  compute_stats: util.ComputeStatsMode = util.ComputeStatsMode.SKIP
  max_examples_per_split: Optional[int] = None
  register_checksums: bool = False
  force_checksums_validation: bool = False
  beam_runner: Optional[Any] = None
  beam_options: Optional[Any] = None
  try_download_gcs: bool = True
  verify_ssl: bool = True
  override_max_simultaneous_downloads: Optional[int] = None
  num_shards: Optional[int] = None
  min_shard_size: int = shard_utils.DEFAULT_MIN_SHARD_SIZE
  max_shard_size: int = shard_utils.DEFAULT_MAX_SHARD_SIZE

  def get_shard_config(self) -> shard_utils.ShardConfig:
    return shard_utils.ShardConfig(
        num_shards=self.num_shards,
        min_shard_size=self.min_shard_size,
        max_shard_size=self.max_shard_size,
    )

  def replace(self, **kwargs: Any) -> DownloadConfig:
    """Returns a copy with updated attributes."""
    return dataclasses.replace(self, **kwargs)


class DownloadManager(object):
  """Manages the download and extraction of files, as well as caching.

  Downloaded files are cached under `download_dir`. The file name of downloaded
   files follows pattern "{sanitized_url}{content_checksum}.{ext}". Eg:
   'cs.toronto.edu_kriz_cifar-100-pythonJDF[...]I.tar.gz'.

  While a file is being downloaded, it is placed into a directory following a
  similar but different pattern:
  "{sanitized_url}{url_checksum}.tmp.{uuid}".

  When a file is downloaded, a "{fname}.INFO.json" file is created next to it.
  This INFO file contains the following information:
  {"dataset_names": ["name1", "name2"],
   "urls": ["http://url.of/downloaded_file"]}

  Extracted files/dirs are stored under `extract_dir`. The file name or
  directory name is the same as the original name, prefixed with the extraction
  method. E.g.
   "{extract_dir}/TAR_GZ.cs.toronto.edu_kriz_cifar-100-pythonJDF[...]I.tar.gz".

  The function members accept either plain value, or values wrapped into list
  or dict. Giving a data structure will parallelize the downloads.

  Example of usage:

  ```
  # Sequential download: str -> str
  train_dir = dl_manager.download_and_extract('https://abc.org/train.tar.gz')
  test_dir = dl_manager.download_and_extract('https://abc.org/test.tar.gz')

  # Parallel download: list -> list
  image_files = dl_manager.download(
      ['https://a.org/1.jpg', 'https://a.org/2.jpg', ...])

  # Parallel download: dict -> dict
  data_dirs = dl_manager.download_and_extract({
     'train': 'https://abc.org/train.zip',
     'test': 'https://abc.org/test.zip',
  })
  data_dirs['train']
  data_dirs['test']
  ```

  For more customization on the download/extraction (ex: passwords, output_name,
  ...), you can pass a `tfds.download.Resource` as argument.
  """

  def __init__(
      self,
      *,
      download_dir: epath.PathLike,
      extract_dir: Optional[epath.PathLike] = None,
      manual_dir: Optional[epath.PathLike] = None,
      manual_dir_instructions: Optional[str] = None,
      url_infos: Optional[Dict[str, checksums.UrlInfo]] = None,
      dataset_name: Optional[str] = None,
      force_download: bool = False,
      force_extraction: bool = False,
      force_checksums_validation: bool = False,
      register_checksums: bool = False,
      register_checksums_path: Optional[epath.PathLike] = None,
      verify_ssl: bool = True,
      max_simultaneous_downloads: Optional[int] = None,
  ):
    """Download manager constructor.

    Args:
      download_dir: Path to directory where downloads are stored.
      extract_dir: Path to directory where artifacts are extracted.
      manual_dir: Path to manually downloaded/extracted data directory.
      manual_dir_instructions: Human readable instructions on how to prepare
        contents of the manual_dir for this dataset.
      url_infos: Urls info for the checksums.
      dataset_name: Name of dataset this instance will be used for. If provided,
        downloads will contain which datasets they were used for.
      force_download: If True, always [re]download.
      force_extraction: If True, always [re]extract.
      force_checksums_validation: If True, raises an error if an URL do not have
        checksums.
      register_checksums: If True, dl checksums aren't checked, but stored into
        file.
      register_checksums_path: Path were to save checksums. Should be set if
        register_checksums is True.
      verify_ssl: `bool`, defaults to True. If True, will verify certificate
        when downloading dataset.
      max_simultaneous_downloads: `int`, optional max number of simultaneous
        downloads.

    Raises:
      FileNotFoundError: Raised if the register_checksums_path does not exist.
    """
    if register_checksums:
      if not register_checksums_path:
        raise ValueError(
            'When register_checksums=True, register_checksums_path should be'
            ' set.'
        )
      register_checksums_path = epath.Path(register_checksums_path)
      if not register_checksums_path.exists():
        # Create the file here to make sure user has write access before
        # starting downloads.
        register_checksums_path.touch()
      else:
        # Make sure the user has write access before downloading any files.
        # (e.g. TFDS installed by admin)
        register_checksums_path.write_text(register_checksums_path.read_text())

    download_dir = epath.Path(download_dir).expanduser()
    if extract_dir:
      extract_dir = epath.Path(extract_dir).expanduser()
    else:
      extract_dir = download_dir / 'extracted'
    if manual_dir:
      manual_dir = epath.Path(manual_dir).expanduser()

    self._download_dir: epath.Path = download_dir
    self._extract_dir: epath.Path = extract_dir
    self._manual_dir: Optional[epath.Path] = (
        manual_dir  # pytype: disable=annotation-type-mismatch  # attribute-variable-annotations
    )
    self._manual_dir_instructions = utils.dedent(manual_dir_instructions)
    self._download_dir.mkdir(parents=True, exist_ok=True)
    self._extract_dir.mkdir(parents=True, exist_ok=True)

    self._force_download = force_download
    self._force_extraction = force_extraction
    self._force_checksums_validation = force_checksums_validation
    self._register_checksums = register_checksums
    self._register_checksums_path = register_checksums_path
    self._verify_ssl = verify_ssl
    self._max_simultaneous_downloads = max_simultaneous_downloads
    self._dataset_name = dataset_name

    # All known URLs: {url: UrlInfo(size=, checksum=)}
    self._url_infos = checksums.get_all_url_infos()
    if url_infos is not None:
      self._url_infos.update(url_infos)

    # To record what is being used: {url: UrlInfo(size, checksum, filename)}
    self._recorded_url_infos: Dict[str, checksums.UrlInfo] = {}
    # These attributes are lazy-initialized since they must be cleared when this
    # object is pickled for Beam. They are then recreated on each worker.
    self.__downloader = None
    self.__extractor = None
    # Executor to avoid blocking other download/extractions when running I/O
    # operations (reading/renaming download file).
    # Only use a single thread as the read/ops are locked by the
    # `build_synchronize_decorator`.
    # Note: This thread is in additions of the download and extraction
    # executors threads.
    self._executor = concurrent.futures.ThreadPoolExecutor(1)

  def __getstate__(self):
    """Remove un-pickleable attributes and return the state."""
    if self._register_checksums:
      # Currently, checksums registration from Beam not supported.
      raise NotImplementedError(
          '`register_checksums` must be disabled in a parallelized '
          'DownloadManager. Please open a PR if you would like this feature.'
      )
    state = self.__dict__.copy()
    state['_DownloadManager__downloader'] = None
    state['_DownloadManager__extractor'] = None
    state['_executor'] = None
    return state

  @property
  def _downloader(self):
    if not self.__downloader:
      self.__downloader = downloader.get_downloader(
          max_simultaneous_downloads=self._max_simultaneous_downloads
      )
    return self.__downloader

  @property
  def _extractor(self):
    if not self.__extractor:
      self.__extractor = extractor.get_extractor()
    return self.__extractor

  @property
  def downloaded_size(self):
    """Returns the total size of downloaded files."""
    return sum(url_info.size for url_info in self._recorded_url_infos.values())

  def _get_dl_path(self, url: str, sha256: str) -> epath.Path:
    return self._download_dir / resource_lib.get_dl_fname(url, sha256)

  @property
  def register_checksums(self):
    """Returns whether checksums are being computed and recorded to file."""
    return self._register_checksums

  @utils.build_synchronize_decorator()
  def _record_url_infos(self):
    """Store in file when recorded size/checksum of downloaded files."""
    checksums.save_url_infos(
        self._register_checksums_path,
        self._recorded_url_infos,
    )

  # Synchronize and memoize decorators ensure same resource will only be
  # processed once, even if passed twice to download_manager.
  @utils.build_synchronize_decorator()
  @utils.memoize()
  def _download(
      self, resource: Union[str, resource_lib.Resource]
  ) -> promise.Promise[epath.Path]:
    """Download resource, returns Promise->path to downloaded file.

    This function:

    1. Reuse cache (`_get_cached_path`) or download the file
    2. Register or validate checksums (`_register_or_validate_checksums`)
    3. Rename download to final path (`_rename_and_get_final_dl_path`)

    Args:
      resource: The URL to download.

    Returns:
      path: The path to the downloaded resource.
    """
    # Normalize the input
    if isinstance(resource, str):
      resource = resource_lib.Resource(url=resource)
    url = resource.url
    assert url is not None, 'URL is undefined from resource.'

    expected_url_info = self._url_infos.get(url)

    # 3 possible destinations for the path:
    # * In `manual_dir` (manually downloaded data)
    # * In `downloads/url_path` (checksum unknown)
    # * In `downloads/checksum_path` (checksum registered)
    manually_downloaded_path = _get_manually_downloaded_path(
        manual_dir=self._manual_dir,
        expected_url_info=expected_url_info,
    )
    url_path = self._get_dl_path(
        url, sha256=hashlib.sha256(url.encode('utf-8')).hexdigest()
    )
    checksum_path = (
        self._get_dl_path(url, sha256=expected_url_info.checksum)
        if expected_url_info
        else None
    )

    # Get the cached path and url_info (if they exists)
    dl_result = _get_cached_path(
        manually_downloaded_path=manually_downloaded_path,
        checksum_path=checksum_path,
        url_path=url_path,
        expected_url_info=expected_url_info,
    )
    if dl_result.path and not self._force_download:  # Download was cached
      logging.info(
          f'Skipping download of {url}: File cached in {dl_result.path}'
      )
      # Still update the progression bar to indicate the file was downloaded
      self._downloader.increase_tqdm(dl_result)
      future = promise.Promise.resolve(dl_result)
    else:
      # Download in an empty tmp directory (to avoid name collisions)
      # `download_tmp_dir` is cleaned-up in `_rename_and_get_final_dl_path`
      dirname = f'{resource_lib.get_dl_dirname(url)}.tmp.{uuid.uuid4().hex}'
      download_tmp_dir = self._download_dir / dirname
      download_tmp_dir.mkdir()
      logging.info(f'Downloading {url} into {download_tmp_dir}...')
      future = self._downloader.download(
          url, download_tmp_dir, verify=self._verify_ssl
      )

    # Post-process the result
    return future.then(
        lambda dl_result: self._register_or_validate_checksums(  # pylint: disable=g-long-lambda
            url=url,
            path=dl_result.path,
            computed_url_info=dl_result.url_info,
            expected_url_info=expected_url_info,
            checksum_path=checksum_path,
            url_path=url_path,
        )
    )

  def _register_or_validate_checksums(
      self,
      path: epath.Path,
      url: str,
      expected_url_info: Optional[checksums.UrlInfo],
      computed_url_info: Optional[checksums.UrlInfo],
      checksum_path: Optional[epath.Path],
      url_path: epath.Path,
  ) -> epath.Path:
    """Validates/records checksums and renames final downloaded path."""
    # `path` can be:
    # * Manually downloaded
    # * (cached) checksum_path
    # * (cached) url_path
    # * `tmp_dir/file` (downloaded path)

    if computed_url_info:
      # Used both in `.downloaded_size` and `_record_url_infos()`
      self._recorded_url_infos[url] = computed_url_info

    if self._register_checksums:
      if not computed_url_info:
        raise ValueError(
            f'Cannot register checksums for {url}: no computed checksum. '
            '--register_checksums with manually downloaded data not supported.'
        )
      # Note:
      # * We save even if `expected_url_info == computed_url_info` as
      #   `expected_url_info` might have been loaded from another dataset.
      # * `register_checksums_path` was validated in `__init__` so this
      #   shouldn't fail.
      self._record_url_infos()

      # Checksum path should now match the new registered checksum (even if
      # checksums were previously registered)
      expected_url_info = computed_url_info
      checksum_path = self._get_dl_path(url, computed_url_info.checksum)
    else:
      # Eventually validate checksums
      # Note:
      # * If path is cached at `url_path` but cached
      #   `computed_url_info != expected_url_info`, a new download has
      #   been triggered (as _get_cached_path returns None)
      # * If path was downloaded but checksums don't match expected, then
      #   the download isn't cached (re-running build will retrigger a new
      #   download). This is expected as it might mean the downloaded file
      #   was corrupted. Note: The tmp file isn't deleted to allow inspection.
      _validate_checksums(
          url=url,
          path=path,
          expected_url_info=expected_url_info,
          computed_url_info=computed_url_info,
          force_checksums_validation=self._force_checksums_validation,
      )

    return self._rename_and_get_final_dl_path(
        url=url,
        path=path,
        expected_url_info=expected_url_info,
        computed_url_info=computed_url_info,
        checksum_path=checksum_path,
        url_path=url_path,
    )

  def _rename_and_get_final_dl_path(
      self,
      url: str,
      path: epath.Path,
      expected_url_info: Optional[checksums.UrlInfo],
      computed_url_info: Optional[checksums.UrlInfo],
      checksum_path: Optional[epath.Path],
      url_path: epath.Path,
  ) -> epath.Path:
    """Eventually rename the downloaded file if checksums were recorded."""
    # `path` can be:
    # * Manually downloaded
    # * (cached) checksum_path
    # * (cached) url_path
    # * `tmp_dir/file` (downloaded path)
    if self._manual_dir and path.is_relative_to(self._manual_dir):
      return path  # Manually downloaded data
    elif path == checksum_path:  # Path already at final destination
      assert computed_url_info == expected_url_info  # Sanity check
      return checksum_path  # pytype: disable=bad-return-type
    elif path == url_path:
      if checksum_path:
        # Checksums were registered: Rename -> checksums_path
        resource_lib.rename_info_file(path, checksum_path, overwrite=True)
        return path.replace(checksum_path)
      else:
        # Checksums not registered: -> do nothing
        return path
    else:  # Path was downloaded in tmp dir
      dst_path = checksum_path or url_path
      resource_lib.write_info_file(
          url=url,
          path=dst_path,
          dataset_name=self._dataset_name,
          original_fname=path.name,
          url_info=computed_url_info,
      )
      path.replace(dst_path)
      path.parent.rmdir()  # Cleanup tmp dir (will fail if dir not empty)
      return dst_path

  @utils.build_synchronize_decorator()
  @utils.memoize()
  def _extract(self, resource: ExtractPath) -> promise.Promise[epath.Path]:
    """Extract a single archive, returns Promise->path to extraction result."""
    if isinstance(resource, epath.PathLikeCls):
      resource = resource_lib.Resource(path=resource)
    path = resource.path
    extract_method = resource.extract_method
    if extract_method == resource_lib.ExtractMethod.NO_EXTRACT:
      logging.info('Skipping extraction for %s (method=NO_EXTRACT).', path)
      return promise.Promise.resolve(path)
    method_name = resource_lib.ExtractMethod(extract_method).name
    extract_path = self._extract_dir / f'{method_name}.{path.name}'
    if not self._force_extraction and extract_path.exists():
      logging.info('Reusing extraction of %s at %s.', path, extract_path)
      return promise.Promise.resolve(extract_path)
    return self._extractor.extract(path, extract_method, extract_path)

  @utils.build_synchronize_decorator()
  @utils.memoize()
  def _download_extract(self, resource):
    """Download-extract `Resource` or url, returns Promise->path."""
    if isinstance(resource, str):
      resource = resource_lib.Resource(url=resource)

    def callback(path):
      resource.path = path
      return self._extract(resource)

    return self._download(resource).then(callback)

  def download_checksums(self, checksums_url):
    """Downloads checksum file from the given URL and adds it to registry."""
    checksums_path = self.download(checksums_url)
    self._url_infos.update(checksums.load_url_infos(checksums_path))

  def download_kaggle_data(self, competition_or_dataset: str) -> epath.Path:
    """Download data for a given Kaggle Dataset or competition.

    Note: This function requires the Kaggle CLI tool.
    Read the installation guide at https://www.kaggle.com/docs/api.

    Args:
      competition_or_dataset: Dataset name (`zillow/zecon`) or competition name
        (`titanic`)

    Returns:
      The path to the downloaded files.
    """
    return kaggle.download_kaggle_data(
        competition_or_dataset, self._download_dir
    )

  @typing.overload
  def download(self, url_or_urls: Url) -> epath.Path:
    ...

  @typing.overload
  def download(self, url_or_urls: Dict[str, Url]) -> Dict[str, epath.Path]:
    ...

  @typing.overload
  def download(self, url_or_urls: Tree[Url]) -> Tree[epath.Path]:
    ...

  def download(self, url_or_urls):
    """Download given url(s).

    Args:
      url_or_urls: url or `list`/`dict` of urls to download and extract. Each
        url can be a `str` or `tfds.download.Resource`.

    Returns:
      downloaded_path(s): `str`, The downloaded paths matching the given input
        url_or_urls.
    """
    # Add progress bar to follow the download state
    with self._downloader.tqdm():
      return _map_promise(self._download, url_or_urls)

  def iter_archive(
      self,
      resource: ExtractPath,
  ) -> Iterator[Tuple[str, typing.BinaryIO]]:
    """Returns iterator over files within archive.

    **Important Note**: caller should read files as they are yielded.
    Reading out of order is slow.

    Args:
      resource: path to archive or `tfds.download.Resource`.

    Returns:
      Generator yielding tuple (path_within_archive, file_obj).
    """
    if isinstance(resource, epath.PathLikeCls):
      resource = resource_lib.Resource(path=resource)
    return extractor.iter_archive(resource.path, resource.extract_method)

  @typing.overload
  def extract(self, path_or_paths: ExtractPath) -> epath.Path:
    ...

  @typing.overload
  def extract(
      self, path_or_paths: Dict[str, ExtractPath]
  ) -> Dict[str, epath.Path]:
    ...

  @typing.overload
  def extract(self, path_or_paths: Tree[ExtractPath]) -> Tree[epath.Path]:
    ...

  def extract(self, path_or_paths):
    """Extract given path(s).

    Args:
      path_or_paths: path or `list`/`dict` of path of file to extract. Each path
        can be a `str` or `tfds.download.Resource`.  If not explicitly specified
        in `Resource`, the extraction method is deduced from downloaded file
        name.

    Returns:
      extracted_path(s): `str`, The extracted paths matching the given input
        path_or_paths.
    """
    # Add progress bar to follow the download state
    with self._extractor.tqdm():
      return _map_promise(self._extract, path_or_paths)

  @typing.overload
  def download_and_extract(self, url_or_urls: Url) -> epath.Path:
    ...

  @typing.overload
  def download_and_extract(
      self, url_or_urls: Dict[str, Url]
  ) -> Dict[str, epath.Path]:
    ...

  @typing.overload
  def download_and_extract(self, url_or_urls: Tree[Url]) -> Tree[epath.Path]:
    ...

  def download_and_extract(self, url_or_urls):
    """Download and extract given url_or_urls.

    Is roughly equivalent to:

    ```
    extracted_paths = dl_manager.extract(dl_manager.download(url_or_urls))
    ```

    Args:
      url_or_urls: url or `list`/`dict` of urls to download and extract. Each
        url can be a `str` or `tfds.download.Resource`.  If not explicitly
        specified in `Resource`, the extraction method will automatically be
        deduced from downloaded file name.

    Returns:
      extracted_path(s): `str`, extracted paths of given URL(s).
    """
    # Add progress bar to follow the download state
    with self._downloader.tqdm():
      with self._extractor.tqdm():
        return _map_promise(self._download_extract, url_or_urls)

  @property
  def download_dir(self) -> epath.Path:
    return self._download_dir

  @functools.cached_property
  def manual_dir(self) -> epath.Path:
    """Returns the directory containing the manually extracted data."""
    if not self._manual_dir:
      raise AssertionError('Manual directory not enabled.')
    if not self._manual_dir_instructions:
      raise ValueError(
          'To access `dl_manager.manual_dir`, please set '
          '`MANUAL_DOWNLOAD_INSTRUCTIONS` in your dataset.'
      )
    if not self._manual_dir.exists() or not list(self._manual_dir.iterdir()):
      raise AssertionError(
          f'Manual directory {self._manual_dir} does not exist or is empty. '
          'Create it and download/extract dataset artifacts in there using '
          f'instructions:\n{self._manual_dir_instructions}'
      )
    return self._manual_dir


def _get_cached_path(
    manually_downloaded_path: Optional[epath.Path],
    checksum_path: Optional[epath.Path],
    url_path: epath.Path,
    expected_url_info: Optional[checksums.UrlInfo],
) -> downloader.DownloadResult:
  """Returns the downloaded path and computed url-info.

  If the path is not cached, or that `url_path` does not match checksums,
  the file will be downloaded again.

  Path can be cached at three different locations:

  Args:
    manually_downloaded_path: Manually downloaded in `dl_manager.manual_dir`
    checksum_path: Cached in the final destination (if checksum known)
    url_path: Cached in the tmp destination (if checksum unknown).
    expected_url_info: Registered checksum (if known)
  """
  # User has manually downloaded the file.
  if manually_downloaded_path and manually_downloaded_path.exists():
    return downloader.DownloadResult(manually_downloaded_path, url_info=None)  # pytype: disable=wrong-arg-types

  # Download has been cached (checksum known)
  elif checksum_path and resource_lib.Resource.exists_locally(checksum_path):
    # `path = f(checksum)` was found, so url_info match
    return downloader.DownloadResult(checksum_path, url_info=expected_url_info)

  # Download has been cached (checksum unknown)
  elif resource_lib.Resource.exists_locally(url_path):
    # Info restored from `.INFO` file
    computed_url_info = _read_url_info(url_path)
    # If checksums are now registered but do not match, trigger a new
    # download (e.g. previous file corrupted, checksums updated)
    if expected_url_info and computed_url_info != expected_url_info:
      return downloader.DownloadResult(path=None, url_info=None)  # pytype: disable=wrong-arg-types
    else:
      return downloader.DownloadResult(url_path, url_info=computed_url_info)

  # Else file not found (or has bad checksums). (re)download.
  else:
    return downloader.DownloadResult(path=None, url_info=None)  # pytype: disable=wrong-arg-types


def _get_manually_downloaded_path(
    manual_dir: Optional[epath.Path],
    expected_url_info: Optional[checksums.UrlInfo],
) -> Optional[epath.Path]:
  """Checks if file is already downloaded in manual_dir."""
  if not manual_dir:  # Manual dir not passed
    return None

  if not expected_url_info or not expected_url_info.filename:
    return None  # Filename unknown.

  manual_path = manual_dir / expected_url_info.filename
  if not manual_path.exists():  # File not manually downloaded
    return None

  return manual_path


def _validate_checksums(
    url: str,
    path: epath.Path,
    computed_url_info: Optional[checksums.UrlInfo],
    expected_url_info: Optional[checksums.UrlInfo],
    force_checksums_validation: bool,
) -> None:
  """Validate computed_url_info match expected_url_info."""
  # If force-checksums validations, both expected and computed url_info
  # should exists
  if force_checksums_validation:
    # Checksum of the downloaded file unknown (for manually downloaded file)
    if not computed_url_info:
      computed_url_info = checksums.compute_url_info(path)
    # Checksums have not been registered
    if not expected_url_info:
      raise ValueError(
          f'Missing checksums url: {url}, yet '
          '`force_checksums_validation=True`. '
          'Did you forget to register checksums?'
      )

  if (
      expected_url_info
      and computed_url_info
      and expected_url_info != computed_url_info
  ):
    msg = (
        f'Artifact {url}, downloaded to {path}, has wrong checksum:\n'
        f'* Expected: {expected_url_info}\n'
        f'* Got: {computed_url_info}\n'
        'To debug, see: '
        'https://www.tensorflow.org/datasets/overview#fixing_nonmatchingchecksumerror'
    )
    raise NonMatchingChecksumError(msg)


def _read_url_info(url_path: epath.PathLike) -> checksums.UrlInfo:
  """Loads the `UrlInfo` from the `.INFO` file."""
  file_info = resource_lib.read_info_file(url_path)
  if 'url_info' not in file_info:
    raise ValueError(
        'Could not find `url_info` in {}. This likely indicates that '
        'the files where downloaded with a previous version of TFDS (<=3.1.0). '
    )
  url_info = file_info['url_info']
  url_info.setdefault('filename', None)
  url_info['size'] = utils.Size(url_info['size'])
  return checksums.UrlInfo(**url_info)


def _map_promise(map_fn, all_inputs):
  """Map the function into each element and resolve the promise."""
  all_promises = tree_utils.map_structure(
      map_fn, all_inputs
  )  # Apply the function
  res = tree_utils.map_structure(
      lambda p: p.get(), all_promises
  )  # Wait promises
  return res



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest
from tensorflow_datasets.core.utils import dtype_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


@pytest.mark.parametrize(
    'input_args,expected_output',
    [
        (np.int64, np.int64),
        (tf.int64, np.int64),
        (np.float64, np.float64),
        (tf.float64, np.float64),
        (tf.string, np.object_),
        (np.uint8, np.uint8),
        (tf.uint8, np.uint8),
    ],
)
def test_tree_parallel_map(input_args, expected_output):
  assert dtype_utils.cast_to_numpy(input_args) == expected_output


@pytest.mark.parametrize(
    'fn,dtype,result',
    [
        (dtype_utils.is_integer, np.int32, True),
        (dtype_utils.is_integer, np.int64, True),
        (dtype_utils.is_integer, np.float32, False),
        (dtype_utils.is_integer, tf.int32, True),
        (dtype_utils.is_integer, tf.int64, True),
        (dtype_utils.is_integer, tf.float32, False),
        (dtype_utils.is_integer, np.uint8, True),
        (dtype_utils.is_integer, tf.uint8, True),
        (dtype_utils.is_bool, np.bool_, True),
        (dtype_utils.is_bool, np.int32, False),
        (dtype_utils.is_bool, tf.bool, True),
        (dtype_utils.is_bool, tf.int32, False),
        (dtype_utils.is_floating, np.float32, True),
        (dtype_utils.is_floating, np.float64, True),
        (dtype_utils.is_floating, np.int32, False),
        (dtype_utils.is_floating, np.int64, False),
        (dtype_utils.is_floating, tf.float32, True),
        (dtype_utils.is_floating, tf.float64, True),
        (dtype_utils.is_floating, tf.int32, False),
        (dtype_utils.is_floating, tf.int64, False),
        (dtype_utils.is_string, np.str_, True),
        (dtype_utils.is_string, np.int32, False),
        (dtype_utils.is_string, tf.string, True),
        (dtype_utils.is_string, tf.int32, False),
        (dtype_utils.is_string, np.uint8, False),
        (dtype_utils.is_string, tf.uint8, False),
    ],
)
def test_dtype(fn, dtype, result):
  assert fn(dtype) == result


def test_is_np_sub_dtype():
  assert dtype_utils.is_np_sub_dtype(np.int32, np.integer)
  assert dtype_utils.is_np_sub_dtype(np.int64, np.integer)
  assert dtype_utils.is_np_sub_dtype(float, np.floating)
  assert not dtype_utils.is_np_sub_dtype(np.int64, np.floating)
  assert not dtype_utils.is_np_sub_dtype(float, np.integer)
  assert dtype_utils.is_np_sub_dtype(np.uint8, np.integer)


def test_is_same_dtype_type():
  assert dtype_utils.is_same_dtype_type(np.int32, tf.int32)
  assert not dtype_utils.is_same_dtype_type(np.int32, np.int64)
  assert dtype_utils.is_same_dtype_type(np.dtype('<U15'), np.dtype('<U0'))



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Custom Translate Dataset."""
import os
from unittest import mock

from tensorflow_datasets.core.folder_dataset import translate_folder
import tensorflow_datasets.public_api as tfds


def _example_dir():
  return os.path.join(
      tfds.testing.test_utils.fake_examples_dir(), 'translate_folder'
  )


original_init = tfds.TranslateFolder.__init__
original_download_and_prepare = tfds.TranslateFolder.download_and_prepare


def new_init(self, root_dir=None, **kwargs):
  assert root_dir is None
  del kwargs
  original_init(self, root_dir=_example_dir())


class TranslateFolderTests(tfds.testing.DatasetBuilderTestCase):
  """Translate dataset tests."""

  DATASET_CLASS = tfds.TranslateFolder
  SPLITS = {
      'train': 4,
      'test': 4,
      'val': 4,
  }

  @classmethod
  def setUpClass(cls):  # pylint:disable = invalid-name
    super().setUpClass()
    super(TranslateFolderTests, cls).setUpClass()
    cls.DATASET_CLASS.__init__ = new_init
    cls.DATASET_CLASS.download_and_prepare = mock.Mock(return_value=None)

  @classmethod
  def tearDownClass(cls):  # pylint:disable = invalid-name
    super().tearDownClass()
    cls.DATASET_CLASS.__init__ = original_init
    cls.DATASET_CLASS.download_and_prepare = original_download_and_prepare

  def test_registered(self):
    # Custom datasets shouldn't be registered
    self.assertNotIn(tfds.TranslateFolder.name, tfds.list_builders())


class TranslateFolderFunctionTest(tfds.testing.TestCase):
  """Tests for TranslateFolder functions."""

  def test_properties(self):
    file_contents = {
        'root_dir/lang1.train.txt': 'line1\nlang1',
        'root_dir/lang2.train.txt': 'line1\nlang2',
        'root_dir/lang3.train.txt': 'line1\nlang3',
        'root_dir/lang1.val.txt': 'line1\nline2\n\nline4',
        'root_dir/lang2.val.txt': 'line1\nline2\n\nline4',
        'root_dir/lang3.val.txt': 'line1\nline2\n\nline4',
        'root_dir/lang1.test.txt': 'line1',
        'root_dir/lang2.test.txt': 'line1',
    }

    with tfds.testing.MockFs() as fs:
      for file in file_contents:
        fs.add_file(file, file_contents[file])

      split_examples, langs = translate_folder._get_split_language_examples(
          'root_dir'
      )
      builder = tfds.TranslateFolder(root_dir='root_dir')

      self.assertEqual(
          split_examples,
          {
              'train': {
                  'lang1': ['line1', 'lang1'],
                  'lang2': ['line1', 'lang2'],
                  'lang3': ['line1', 'lang3'],
              },
              'val': {
                  'lang1': ['line1', 'line2', '', 'line4'],
                  'lang2': ['line1', 'line2', '', 'line4'],
                  'lang3': ['line1', 'line2', '', 'line4'],
              },
              'test': {
                  'lang1': ['line1'],
                  'lang2': ['line1'],
              },
          },
      )

      self.assertEqual(builder.info.splits['train'].num_examples, 2)
      self.assertEqual(builder.info.splits['val'].num_examples, 4)
      self.assertEqual(builder.info.splits['test'].num_examples, 1)

      expected_languages = [
          'lang1',
          'lang2',
          'lang3',
      ]

      self.assertEqual(expected_languages, langs)
      self.assertEqual(list(builder.info.features.keys()), expected_languages)


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library functions and types for the robomimic datasets."""

import dataclasses
import enum
from typing import Any, Dict, List, Mapping

from etils import epath
import h5py
import numpy as np
import tensorflow_datasets.public_api as tfds
import tree



class Task(enum.Enum):
  LIFT = 'lift'
  CAN = 'can'
  SQUARE = 'square'
  TRANSPORT = 'transport'
  TOOL_HANG = 'tool_hang'

  def __str__(self) -> str:
    return self.value


class DataSource(enum.Enum):
  PH = 'ph'  # proficient human, i.e. expert demonstrations
  MH = 'mh'  # mixed-human, i.e. more suboptimal than proficient human
  MG = 'mg'  # machine generated, i.e. soft actor critic with dense reward

  def __str__(self) -> str:
    return self.value


class ObservationType(enum.Enum):
  IMAGE = 'image'  # image-based observations (i.e. exteroception)
  LOW_DIM = 'low_dim'  # internal state-based observations (i.e. proprioception)

  def __str__(self) -> str:
    return self.value


@dataclasses.dataclass
class BuilderConfig(tfds.core.BuilderConfig):
  """Configuration of the dataset versions."""

  task: Task = Task.LIFT
  dataset: DataSource = DataSource.PH
  filename: ObservationType = ObservationType.LOW_DIM
  horizon: int = 100


TASKS = {
    Task.LIFT: {
        'datasets': [DataSource.PH, DataSource.MH, DataSource.MG],
        'object': 10,
        'states': 32,
        'horizon': 400,
        'action_size': 7,
    },
    Task.CAN: {
        'datasets': [DataSource.PH, DataSource.MH, DataSource.MG],
        'object': 14,
        'states': 71,
        'horizon': 400,
        'action_size': 7,
    },
    Task.SQUARE: {
        'datasets': [DataSource.PH, DataSource.MH],
        'object': 14,
        'states': 45,
        'horizon': 400,
        'action_size': 7,
    },
    Task.TRANSPORT: {
        'datasets': [DataSource.PH, DataSource.MH],
        'object': 41,
        'states': 115,
        'horizon': 700,
        'action_size': 14,
    },
    Task.TOOL_HANG: {
        'datasets': [
            DataSource.PH,
        ],
        'object': 44,
        'states': 58,
        'horizon': 700,
        'action_size': 7,
    },
}


def make_builder_configs(dataset: DataSource):
  """Creates the PH build configs."""
  configs = []
  for task, details in TASKS.items():
    if dataset in details['datasets']:
      for observation_type in [ObservationType.IMAGE, ObservationType.LOW_DIM]:
        # pytype: disable=wrong-keyword-args
        configs.append(
            # name is inherited from base dataclass unbeknownst to the linter
            BuilderConfig(  # pylint: disable=unexpected-keyword-arg
                name=f'{task}_{dataset}_{observation_type}',
                task=task,
                dataset=dataset,
                filename=observation_type,
                horizon=details['horizon'],
            )
        )
        # pytype: enable=wrong-keyword-args
  return configs


def _concat_obs(base_obs, extra_obs):
  return np.append(base_obs, np.expand_dims(extra_obs, 0), axis=0)


def episode_metadata(
    mask: Mapping[str, List[str]], episode_key: str
) -> Dict[str, bool]:
  """Builds the metadata of an episode.

  Args:
    mask: A dict that maps flags to the list of episodes for which this flag is
      true.
    episode_key: The key of an episode.

  Returns:
    Dictionary that maps the flags to the value corresponding to the given
    episode.
  """
  metadata = {}
  for k in mask:
    metadata[k] = episode_key in mask[k]
  return metadata


def build_episode(steps: Mapping[str, Any]) -> Dict[str, Any]:
  """Builds a full episode dict.

  Args:
    steps: A dict with each key corresponding to an episode of that variable.

  Returns:
   A dict with data specific to one episode.
  """

  # This is the base episode length without the end step, that step will get
  # added manually.  `[:]` notation is necessary to extract data from the hdf5
  # object.
  ep_length = steps['actions'][:].shape[0]
  episode = {}
  episode['is_first'] = np.append(True, np.zeros(ep_length, dtype=bool))

  # The standard 'obs' needs to be extended with the last element from
  # 'next_obs' to reconstruct the full sequence.
  obs = tree.map_structure(np.array, dict(steps['obs']))
  last_obs = tree.map_structure(lambda el: el[-1], dict(steps['next_obs']))
  concat_obs = tree.map_structure(_concat_obs, obs, last_obs)

  actions = steps['actions'][:]
  dones = steps['dones'][:]
  episode['observation'] = concat_obs
  episode['action'] = np.append(
      actions, np.expand_dims(np.zeros(actions[0].shape), 0), axis=0
  )
  episode['reward'] = np.append(steps['rewards'][:], 0)
  episode['discount'] = np.append(np.ones(ep_length), 0)
  # Offset `dones` by one as it corresponds to the `next_obs`:
  # "done signal, equal to 1 if playing the corresponding action in the state
  # should terminate the episode".  Since playing another action may not end the
  # episode, it is actually representing whether the following state is
  # the terminal state.
  episode['is_terminal'] = np.append(False, dones).astype(bool)
  episode['is_last'] = np.append(np.zeros(ep_length, dtype=bool), True)
  states = steps['states'][:]
  episode['states'] = np.append(
      states, np.expand_dims(np.zeros(states[0].shape), 0), axis=0
  )

  return episode


def tensor_feature(size: int, doc=None) -> tfds.features.Tensor:
  return tfds.features.Tensor(
      shape=(size,),
      dtype=np.float64,
      encoding=tfds.features.Encoding.ZLIB,
      doc=doc,
  )


def image_feature(size: int, doc=None) -> tfds.features.Image:
  return tfds.features.Image(
      shape=(size, size, 3), dtype=np.uint8, encoding_format='png', doc=doc
  )


class RobomimicBuilder(tfds.core.GeneratorBasedBuilder, skip_registration=True):
  """DatasetBuilder for robomimic datasets."""

  VERSION: tfds.core.Version
  RELEASE_NOTES: Dict[str, str]
  BUILDER_CONFIGS: List[tfds.core.BuilderConfig]
  DATASET_NAME: str
  DATASET_FILE_EXTENSION: str = ''


  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=self._get_features(),
        supervised_keys=None,
        homepage='https://arise-initiative.github.io/robomimic-web/',
    )

  def _get_features(self) -> tfds.features.FeaturesDict:
    obs_dim = TASKS[self.builder_config.task]['object']
    states_dim = TASKS[self.builder_config.task]['states']
    action_size = TASKS[self.builder_config.task]['action_size']

    observation = {
        'object': tensor_feature(
            obs_dim,
        ),
        'robot0_eef_pos': tensor_feature(3, doc='End-effector position'),
        'robot0_eef_quat': tensor_feature(4, doc='End-effector orientation'),
        'robot0_eef_vel_ang': tensor_feature(
            3, doc='End-effector angular velocity'
        ),
        'robot0_eef_vel_lin': tensor_feature(
            3, doc='End-effector cartesian velocity'
        ),
        'robot0_gripper_qpos': tensor_feature(2, doc='Gripper position'),
        'robot0_gripper_qvel': tensor_feature(2, doc='Gripper velocity'),
        'robot0_joint_pos': tensor_feature(7, doc='7DOF joint positions'),
        'robot0_joint_pos_cos': tensor_feature(7),
        'robot0_joint_pos_sin': tensor_feature(7),
        'robot0_joint_vel': tensor_feature(7, doc='7DOF joint velocities'),
    }
    if self.builder_config.task == Task.TRANSPORT:
      observation['robot1_eef_pos'] = tensor_feature(
          3, doc='End-effector position'
      )
      observation['robot1_eef_quat'] = tensor_feature(
          4, doc='End-effector orientation'
      )
      observation['robot1_eef_vel_ang'] = tensor_feature(
          3, doc='End-effector angular velocity'
      )
      observation['robot1_eef_vel_lin'] = tensor_feature(
          3, doc='End-effector cartesian velocity'
      )
      observation['robot1_gripper_qpos'] = tensor_feature(
          2, doc='Gripper position'
      )
      observation['robot1_gripper_qvel'] = tensor_feature(
          2, doc='Gripper velocity'
      )
      observation['robot1_joint_pos'] = tensor_feature(
          7, doc='7DOF joint positions'
      )
      observation['robot1_joint_pos_cos'] = tensor_feature(7)
      observation['robot1_joint_pos_sin'] = tensor_feature(7)
      observation['robot1_joint_vel'] = tensor_feature(
          7, doc='7DOF joint velocities'
      )

    if self.builder_config.filename == ObservationType.IMAGE:
      if self.builder_config.task == Task.TOOL_HANG:
        observation['robot0_eye_in_hand_image'] = image_feature(240)
        observation['sideview_image'] = image_feature(240)
      elif self.builder_config.task == Task.TRANSPORT:
        observation['robot0_eye_in_hand_image'] = image_feature(84)
        observation['robot1_eye_in_hand_image'] = image_feature(84)
        observation['shouldercamera0_image'] = image_feature(84)
        observation['shouldercamera1_image'] = image_feature(84)
      else:
        observation['agentview_image'] = image_feature(84)
        observation['robot0_eye_in_hand_image'] = image_feature(84)

    # metadata depends on the quality type
    metadata = self._get_metadata()

    features = tfds.features.FeaturesDict({
        'horizon': np.int32,
        'episode_id': np.str_,
        'steps': tfds.features.Dataset({
            'action': tensor_feature(action_size),
            'observation': observation,
            'reward': np.float64,
            'is_first': np.bool_,
            'is_last': np.bool_,
            'is_terminal': np.bool_,
            'discount': np.int32,
            'states': tensor_feature(states_dim),
        }),
        **metadata,
    })
    return features

  def _get_metadata(self) -> Dict[Any, Any]:
    return {}

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    # Machine generated datasets have an additional variant as they are trained
    # using dense rather than sparse rewards. We currently are only interested
    # in the sparse rewards, for consistency with the other datasets.
    ext = self.DATASET_FILE_EXTENSION
    filepath = (
        'http://downloads.cs.stanford.edu/downloads/rt_benchmark/'
        f'{self.builder_config.task}/{self.builder_config.dataset}/'
        f'{self.builder_config.filename}{ext}.hdf5'
    )
    path = dl_manager.download_and_extract({'file_path': filepath})
    return {
        'train': self._generate_examples(path),
    }

  def _generate_examples(self, path):
    """Yields examples."""
    path = epath.Path(path['file_path'])
    with path.open('rb') as f:
      with h5py.File(f, 'r') as dataset_file:
        data = dataset_file['data']
        if 'mask' in dataset_file.keys():
          mask = dataset_file['mask']
          string_mask = {}
          for k in mask:
            string_mask[k] = []
            for element in mask[k]:
              string_mask[k].append(element.decode('UTF-8'))
          mask = string_mask
          for key in data:
            yield key, {
                'steps': build_episode(data[key]),
                'horizon': self.builder_config.horizon,
                'episode_id': key,
                **episode_metadata(mask, key),
            }
        else:
          for key in data:
            yield key, {
                'steps': build_episode(data[key]),
                'horizon': self.builder_config.horizon,
                'episode_id': key,
            }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""rlu_control_suite dataset."""

from __future__ import annotations

import dataclasses
from typing import Any, Dict, Optional

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rl_unplugged import rlu_common

_CONTROL_SUITE_DESCRIPTION = """
DeepMind Control Suite [Tassa et al., 2018](https://arxiv.org/abs/1801.00690)
is a set of control tasks implemented in MuJoCo
[Todorov et al., 2012](https://homes.cs.washington.edu/~todorov/papers/TodorovIROS12.pdf).
We consider a subset of the tasks provided in the suite that cover a wide range
of difficulties.

Most of the datasets in this domain are generated using D4PG. For the
environments Manipulator insert ball and Manipulator insert peg we use V-MPO
[Song et al., 2020](https://arxiv.org/abs/1909.12238) to generate the data as
D4PG is unable to solve these tasks. We release datasets for 9 control suite
tasks. For details on how the dataset was generated, please refer to the paper.

DeepMind Control Suite is a traditional continuous action RL benchmark.
In particular, we recommend you test your approach in DeepMind Control Suite if
you are interested in comparing against other state of the art offline RL
methods.
"""

_CITATION = """
@inproceedings{gulcehre2020rl,
 title = {RL Unplugged: A Suite of Benchmarks for Offline Reinforcement Learning},
 author = {Gulcehre, Caglar and Wang, Ziyu and Novikov, Alexander and Paine, Thomas and G\'{o}mez, Sergio and Zolna, Konrad and Agarwal, Rishabh and Merel, Josh S and Mankowitz, Daniel J and Paduraru, Cosmin and Dulac-Arnold, Gabriel and Li, Jerry and Norouzi, Mohammad and Hoffman, Matthew and Heess, Nicolas and de Freitas, Nando},
 booktitle = {Advances in Neural Information Processing Systems},
 pages = {7248--7259},
 volume = {33},
 year = {2020}
}
"""


@dataclasses.dataclass
class ControlSuiteBuilderConfig(tfds.core.BuilderConfig):
  observation_size: Optional[Dict[str, int]] = None
  action_size: int = 0


# pytype: disable=wrong-keyword-args
_BUILDER_CONFIGS = [
    ControlSuiteBuilderConfig(
        name='cartpole_swingup',
        observation_size={'position': 3, 'velocity': 2},
        action_size=1,
    ),
    ControlSuiteBuilderConfig(
        name='cheetah_run',
        observation_size={'position': 8, 'velocity': 9},
        action_size=6,
    ),
    ControlSuiteBuilderConfig(
        name='finger_turn_hard',
        observation_size={
            'position': 4,
            'velocity': 3,
            'target_position': 2,
            'dist_to_target': 1,
        },
        action_size=2,
    ),
    ControlSuiteBuilderConfig(
        name='fish_swim',
        observation_size={
            'target': 3,
            'velocity': 13,
            'upright': 1,
            'joint_angles': 7,
        },
        action_size=5,
    ),
    ControlSuiteBuilderConfig(
        name='humanoid_run',
        observation_size={
            'velocity': 27,
            'com_velocity': 3,
            'torso_vertical': 3,
            'extremities': 12,
            'head_height': 1,
            'joint_angles': 21,
        },
        action_size=21,
    ),
    ControlSuiteBuilderConfig(
        name='manipulator_insert_ball',
        observation_size={
            'arm_pos': 16,
            'arm_vel': 8,
            'touch': 5,
            'hand_pos': 4,
            'object_pos': 4,
            'object_vel': 3,
            'target_pos': 4,
        },
        action_size=5,
    ),
    ControlSuiteBuilderConfig(
        name='manipulator_insert_peg',
        observation_size={
            'arm_pos': 16,
            'arm_vel': 8,
            'touch': 5,
            'hand_pos': 4,
            'object_pos': 4,
            'object_vel': 3,
            'target_pos': 4,
        },
        action_size=5,
    ),
    ControlSuiteBuilderConfig(
        name='walker_stand',
        observation_size={
            'orientations': 14,
            'velocity': 9,
            'height': 1,
        },
        action_size=6,
    ),
    ControlSuiteBuilderConfig(
        name='walker_walk',
        observation_size={
            'orientations': 14,
            'velocity': 9,
            'height': 1,
        },
        action_size=6,
    ),
]
# pytype: enable=wrong-keyword-args


def _sequence_feature(
    size: Optional[int] = None,
) -> tf.io.FixedLenSequenceFeature:
  if size:
    shape = [size]
  else:
    shape = []
  return tf.io.FixedLenSequenceFeature(
      shape, dtype=tf.float32, allow_missing=True
  )


class RluControlSuite(rlu_common.RLUBuilder):
  """DatasetBuilder for rlu_control_suite dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  BUILDER_CONFIGS = _BUILDER_CONFIGS

  _SHARDS = 100
  _INPUT_FILE_PREFIX = 'gs://rl_unplugged/dm_control_suite_episodes/'

  def get_features_dict(self):
    return tfds.features.FeaturesDict({
        'steps': tfds.features.Dataset({
            'observation': {
                k: rlu_common.float_tensor_feature(v)
                for k, v in self.builder_config.observation_size.items()
            },
            'action': tfds.features.Tensor(
                shape=(self.builder_config.action_size,), dtype=np.float32
            ),
            'reward': np.float32,
            'is_terminal': np.bool_,
            'is_first': np.bool_,
            'is_last': np.bool_,
            'discount': np.float32,
        }),
        'episode_id': np.int64,
        'timestamp': np.int64,
    })

  def get_description(self):
    return _CONTROL_SUITE_DESCRIPTION

  def get_citation(self):
    return _CITATION

  def get_file_prefix(self):
    task = self.builder_config.name
    return f'{self._INPUT_FILE_PREFIX}/{task}/train'

  def num_shards(self):
    return self._SHARDS

  def _get_example_specs(self):
    obs_features = {
        f'observation/{k}': _sequence_feature(v)
        for k, v in self.builder_config.observation_size.items()
    }
    return {
        **obs_features,
        'action': _sequence_feature(self.builder_config.action_size),
        'discount': _sequence_feature(),
        'reward': _sequence_feature(),
        'step_type': _sequence_feature(),
        'episode_id': tf.io.FixedLenFeature([], tf.int64),
        'timestamp': tf.io.FixedLenFeature([], tf.int64),
    }

  def tf_example_to_step_ds(
      self, tf_example: tf.train.Example
  ) -> Dict[str, Any]:
    feature_description = self._get_example_specs()

    data = tf.io.parse_single_example(tf_example, feature_description)
    episode_length = tf.size(data['discount'])
    is_first = tf.concat(
        [[True], [False] * tf.ones(episode_length - 1)], axis=0
    )
    is_last = tf.concat([[False] * tf.ones(episode_length - 1), [True]], axis=0)
    is_terminal = [False] * tf.ones(episode_length, tf.int64)

    # The data is in RSA alignment, we realign it to SAR to comply with the
    # RLDS standard.
    discount = data['discount'][1:]
    if discount[-1] == 0.0:
      is_terminal = tf.concat(
          [[False] * tf.ones(episode_length - 1, tf.int64), [True]], axis=0
      )
      # If the episode ends in a terminal state, in the last step only the
      # observation has valid information (the terminal state).
      discount = tf.concat([discount, [0.0]], axis=0)
    else:
      discount = tf.concat([discount, [1.0]], axis=0)

    reward = tf.concat([data['reward'][1:], [0.0]], axis=0)

    obs_prefix = 'observation/'
    obs = {}
    for k in feature_description:
      if k.startswith(obs_prefix):
        new_k = k[len(obs_prefix) :]
        obs[new_k] = data[k]
    episode = {
        'steps': {
            'observation': obs,
            'action': data['action'],
            'reward': reward,
            'discount': discount,
            'is_first': is_first,
            'is_last': is_last,
            'is_terminal': is_terminal,
        },
        'episode_id': data['episode_id'],
        'timestamp': data['timestamp'],
    }
    return episode



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Documentation utils for the website.

Note: To be detected, doc decorators should be applied between descriptors
and other decorators.

```py
class A:

  @staticmethod
  @tfds.core.utils.docs.deprecated
  @other_decorator
  def f():
    pass
```

The functions exposed below are dummy decorators. This allows not having to load
TensorFlow. The functions are monkey patched when needed in
scripts/documentation/build_api_docs.py with actual TensorFlow documentation
decorators.
"""

from typing import Any, TypeVar

_T = TypeVar('_T')


# Some TensorFlow documentation decorators used to obfuscate typing: b/262340871
def _no_op_decorator_without_type(obj) -> Any:
  return obj


def _no_op_decorator(obj: _T) -> _T:
  return obj


deprecated = _no_op_decorator_without_type
doc_private = _no_op_decorator
do_not_doc = _no_op_decorator
do_not_doc_inheritable = _no_op_decorator
do_not_doc_in_subclasses = _no_op_decorator_without_type



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for conll_dataset_builder."""
import textwrap
from unittest import mock

from etils import epath
import pytest
from tensorflow_datasets.core.dataset_builders.conll import conll_dataset_builder
from tensorflow_datasets.core.dataset_builders.conll import conll_dataset_builder_utils as conll_lib
import tensorflow_datasets.public_api as tfds

_FOLDER_PATH = "mock/path"

_VALID_INPUT = textwrap.dedent(
    """
-DOCSTART- -X- -X- O
Winter NN B-NP O
is VBZ B-VP O

Air NN I-NP O
. . O O
"""
)

_INVALID_INPUT = textwrap.dedent(
    """
Winter NN B-NP
is VBZ B-VP O

Air NN I-NP O
. . O O
"""
)

_INPUT_PATH = epath.Path(_FOLDER_PATH, "input_path.txt")


class DummyConllDataset(conll_dataset_builder.ConllDatasetBuilder):
  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {"1.0.0": "Dummy notes."}
  BUILDER_CONFIGS = [conll_lib.CONLL_2003_CONFIG]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.create_dataset_info(
        description="Dummy CoNLL dataset.",
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    del dl_manager
    return {"train": self._generate_examples(_INPUT_PATH)}


def test_generate_example():
  tf_mock = mock.Mock()
  tf_mock.gfile.GFile.return_value = _VALID_INPUT
  expected_examples = []

  dataset = DummyConllDataset()

  with tfds.testing.MockFs() as fs:
    fs.add_file(path=_INPUT_PATH, content=_VALID_INPUT)
    examples = list(dataset._generate_examples(_INPUT_PATH))

    expected_examples = [
        (
            0,
            {
                "tokens": ["Winter", "is"],
                "pos": ["NN", "VBZ"],
                "chunks": ["B-NP", "B-VP"],
                "ner": ["O", "O"],
            },
        ),
        (
            1,
            {
                "tokens": ["Air", "."],
                "pos": ["NN", "."],
                "chunks": ["I-NP", "O"],
                "ner": ["O", "O"],
            },
        ),
    ]

    assert examples == expected_examples

    for _, example in examples:
      assert len(example) == len(conll_lib.CONLL_2003_ORDERED_FEATURES)

  assert len(examples) == 2


def test_generate_corrupted_example():
  tf_mock = mock.Mock()
  tf_mock.gfile.GFile.return_value = _VALID_INPUT
  dataset = DummyConllDataset()

  error_line = "Winter NN B-NP"
  error_msg = (
      f"Mismatch in the number of features found in line: {error_line}\n\n"
      "Should be 4, but found 3"
  )
  with pytest.raises(ValueError, match=error_msg):
    with tfds.testing.MockFs() as fs:
      fs.add_file(path=_INPUT_PATH, content=_INVALID_INPUT)
      list(dataset._generate_examples(_INPUT_PATH))



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Image feature."""

from __future__ import annotations

from typing import Dict, List, Optional, Union

from etils import epath
from tensorflow_datasets.core.features import class_label_feature
from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features import image_feature
from tensorflow_datasets.core.proto import feature_pb2
from tensorflow_datasets.core.utils import type_utils

Json = type_utils.Json

_LabelArg = Union[List[str], epath.PathLike, None, int]


class LabeledImage(image_feature.Image):
  """Image with additional label metadata.

  This feature connector is similar to `tfds.features.Image`, but expose
  additional attributes from `tfds.features.ClassLabel` to access label
  metadata.

  This can be used for segmentation tasks were each pixel is labeled.

  ```python
  builder = tfds.builder('my_dataset')
  builder.info.features['label_image'].names == ['background', 'car', ...]
  ```
  """

  def __init__(
      self,
      *,
      labels: _LabelArg,
      shape: Optional[type_utils.Shape] = None,
      dtype: Optional[type_utils.TfdsDType] = None,
      encoding_format: Optional[str] = None,
      doc: feature_lib.DocArg = None,
  ):
    """Constructor.

    Labels can be defined as:

    * `List[str]` (e.g. `['background', 'car', 'truck', ...]`): The list of
      label strings.
    * `PathLike` (e.g. `'/path/to/label.txt'`): A file containing the labels
      (one per-line).
    * `int` (e.g. `12`): The total number of labels, if the actual label names
      string are unknown (discouraged).
    * `None`: If the number of label is unknown (discouraged).

    Args:
      labels: Labels metadata (see docstring above)
      shape: Image shape (see `tfds.features.Image.__init__`)
      dtype: Image dtype (see `tfds.features.Image.__init__`)
      encoding_format: 'jpeg' or 'png' (see `tfds.features.Image.__init__`)
      doc: Documentation of this feature (e.g. description).
    """
    super().__init__(
        # Label images have a single channel
        shape=shape or (None, None, 1),
        dtype=dtype,
        encoding_format=encoding_format,
        use_colormap=True,  # LabeledImage always use colormap
        doc=doc,
    )
    if self.shape[-1] != 1:
      raise ValueError(
          f'LabeledImage shape should have a single channel. Got: {shape}'
      )
    label_kwargs = _labels_to_kwarg(labels)
    self._class_label = class_label_feature.ClassLabel(**label_kwargs)

  @property
  def num_classes(self) -> Optional[int]:
    return self._class_label.num_classes

  @property
  def names(self) -> List[str]:
    return self._class_label.names

  def save_metadata(self, data_dir, feature_name=None) -> None:
    super().save_metadata(data_dir=data_dir, feature_name=feature_name)
    self._class_label.save_metadata(
        data_dir=data_dir, feature_name=feature_name
    )

  def load_metadata(self, data_dir, feature_name=None) -> None:
    super().load_metadata(data_dir=data_dir, feature_name=feature_name)
    self._class_label.load_metadata(
        data_dir=data_dir, feature_name=feature_name
    )

  def _additional_repr_info(self):
    return {'num_classes': self.num_classes}

  @classmethod
  def from_json_content(
      cls, value: Union[Json, feature_pb2.ImageFeature]
  ) -> 'LabeledImage':
    if isinstance(value, dict):
      # For backwards compatibility
      return cls(**value)
    return cls(
        shape=feature_lib.from_shape_proto(value.shape),
        dtype=feature_lib.dtype_from_str(value.dtype),
        encoding_format=value.encoding_format or None,
        labels=value.label.num_classes or None,
    )

  def to_json_content(self) -> feature_pb2.ImageFeature:
    feature = super().to_json_content()
    feature.label.CopyFrom(feature_pb2.ClassLabel(num_classes=self.num_classes))
    return feature


def _labels_to_kwarg(labels: _LabelArg) -> Dict[str, _LabelArg]:
  """Creates the `ClassLabel.__init__` kwargs."""
  if labels is None:
    return {}
  elif isinstance(labels, int):
    kwarg_name = 'num_classes'
  elif isinstance(labels, epath.PathLikeCls):
    kwarg_name = 'names_file'
  elif isinstance(labels, list):
    kwarg_name = 'names'
  else:
    raise TypeError(
        f'Invalid `labels` type: {type(labels)}. Should be one of '
        'list[labels], path, num_labels'
    )
  return {kwarg_name: labels}
